#******************************************************************************
# 
# File   : argosCC.py
# Authors:	Clay Shepard (cws [at] rice.edu)
# License:	Copyright 2015, Rice Efficient Computing Group. All rights reserved.
#
#******************************************************************************

import sys
import asyncio
import struct
import numpy as np
import time
import configparser 
import os
from argoscc.warpNode import *
from argoscc.argos_init import *
#from argoscc.argos_calibrate import *
import math
import h5py
import time
import datetime
import logging
import socket
from argoscc.interface_lms6 import InterfaceLMS6
from argoscc.interface_m2829 import InterfaceM2829

try:
	from argoscc.argos_hdf5_clib import *
except ImportError:
	print("Warning! Unable to find Argos_HDF5_CLIB.  Using slower Python implementation.")
	
#TODO: rename file to argoscc.py to follow python convention.

class ArgosCC:
	"""
	The Argos Central Controller.
	
	The central controller is a high level controller which manages all of the nodes, including gain control, channel collection, and beamformer computation.
	Accelerated C libraries are available for channel collection and turbo decoding.

	Note: There must be at least one bs_node (the master) and one mobile node.
	Note: Fedora's firewall blocks the reported CSI by default, so it must be disabled, or configured to allow it (e.g. port 55104, or set to internal).
	Note: In linux you have to add "net.core.rmem_max = 67108864" to /etc/sysctl.conf then do a sysctl -p to load the changes	
	Note: In linux to enable jumbo frames do something like "sudo ifconfig eth0 mtu 9000"
	Note: In Windows to enable jumbo frames you must run:

		netsh interface ipv4 set subinterface "Local Area Connection" mtu=9100 store=persistent
		netsh interface ipv4 show subinterfaces
		
		To make it persist: click start type regedit the navigate to [HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services\Tcpip\Parameters\Interfaces\[Adapter ID]]
	"""
		
	class UplinkDataReceiver:
		"""The UplinkDataReceiver handles all packets sent by the WarpNodes that contain the uplink IQ samples -- it simply parses them out and saves them in the HDF5 file."""
		def __init__(self, cc, host_addr='0.0.0.0', port=52525):
			self._host_addr = host_addr
			self._host_port = port	
			self.cc = cc
			self.logger = cc.logger
			self.loop = asyncio.new_event_loop()
			self.loop.run_until_complete(self.loop.create_datagram_endpoint(lambda: self, local_addr=(self._host_addr,self._host_port) ) )
			self.socket = self.transport.get_extra_info('socket')

			sock_buf_size = int(1024*1024*64*2) #macs complain about a float...
			try:
				self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, sock_buf_size)  #this is critical for not dropping packets! #Note! In linux you have to add "net.core.rmem_max = 67108864" to /etc/sysctl.conf then do a sysctl -p to load the changes
			except:
				print("Warning!  There was a problem setting the receive socket buffer size.  This happens on Macs.  Be wary of dropped packets!  It should be %d, but is actually %d" % (sock_buf_size, self.socket.getsockopt(socket.SOL_SOCKET,socket.SO_RCVBUF)))
				self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 1024*1024*6*2) #todo check what actually fails on macs

			self.thread_loop = threading.Thread(target=self.loop_in_thread)
			self.thread_loop.start()	

			self.counter = 0 #temporary metric
		
		def datagram_received(self,data,addr):
			""" 
			This is what handles all uplink automatically sent by the nodes (e.g. uplink IQ, but not any command responses, which are handled by the WarpNode object.)
			
			Currently, if the 'dump_up_buff' command is sent to the bs_nodes with the number of samples to dump and port they will start reporting uplink IQ to the specified MAC/IP (set by set_cc_mac_ip).
			The packet format is FrameNumber RADIO_1, RADIO_2, RADIO_3, RADIO_4.
			All values are reported as 4 byte IQ samples.			
			
			This function parses those packets and saves them in the circular buffers.
			Once all nodes have reported for a given frame, the frame is recorded to disk by copying them to the h5py file.
			
			Note that this was intially designed to be multithreaded to improve performance, but the Python GIL makes multithreading essentially worthless.
			TODO: In the future, the recording to disk should only happen in chunks, not every frame.
			"""
			warp_header_len = 12
			csi_start = warp_header_len + 6 #2 bytes pad at beginning + sizeof(int), for frame_number
			#self.logger.info("data received from %s" % addr[0])
			#we're only sending the transport header, not the command/response header #http://warpproject.org/trac/wiki/WARPLab/Reference/Architecture/WireFormat
			header = struct.unpack(">xx2H xB 3H", data[0:14])  #> means big endian.  H is 2 bytes, B is 1, I is 3, x is disregarded single pad byte.  capital is unsigned
			ID = header[0]
			nodeID = header[1]
			pkt_type = header[2]
			length = header[3]
			seq_num = header[4]
			flags = header[5] 
	
			frame_number = struct.unpack("I", data[csi_start-4:csi_start])[0]
			frame_idx = frame_number % self.cc._num_frames  #index in to our cache arrays (circular buffers)			
			bs_ant_idx = struct.unpack("I", data[csi_start:csi_start+4])[0]-1 #the second arg is the antenna idx for consistency it starts with 1, so we have to subtract 1
				
			#print("Packet from node %d received!  Frame number %d, length %d. " % (nodeID, frame_number, len(data) ) )
			#self.logger.info("Packet from node %x with cmd %d (cmdID %d, cmdGrp %d) and %d cmdArgs received!  Frame number %d, length %d. " % (nodeID, cmd, cmdID, cmdGrp, numArgs, frame_number, len(data) ) )
			csi_len = self.cc.UplinkData.shape[2]*4 #4bytes per sample (frame, bsant, samples, iq)		
		
			#TODO: check to make sure frame number isn't outside of circular buffer.
		
			#todo: fix this... (check length of packet)
			#if len(data) != :  #2 bytes padding at beginning and end
			#	print("Warning! Data length mismatch received pack of length %d, should be %d.  WARP header reported %d, should be %d  Discarding packet." % (len(data), csi_start+csi_len + rssi_len + 2, length, (len(data)-warp_header_len) ) ) #self.cc.logger.warn
			#	return
			
			if nodeID not in self.cc.BSID2idx:
				self.cc.logger.warn("Warning! UplinkData Received data from bogus nodeID %d.  Discarding packet." % nodeID)
				return
			bs_ant_idx += cc.BSID2idx[nodeID]*cc._num_ant_per_bs	#each node only reports its internal antenna ID, we want the ID on the array
			
			CSI_raw = np.fromstring(data[csi_start:csi_start+csi_len], dtype='i2') #use '>i2' for big-endian
			CSI_Q = CSI_raw[0::2]  #these are reversed because of the endian swap.
			CSI_I = CSI_raw[1::2]
			
			self.cc.UplinkData_buffer[frame_idx,bs_ant_idx,:,0] = CSI_I
			self.cc.UplinkData_buffer[frame_idx,bs_ant_idx,:,1] = CSI_Q

			self.counter += 1
			
			with self.cc.num_reported_uplink_lock: #we could create _num_frames locks to avoid excessive contention, but it shouldn't help that much if each frame is processed in realtime
				self.cc.num_reported_uplink[frame_idx] += 1  #notify that one more node reported in
				mynum_reported = self.cc.num_reported_uplink[frame_idx] #make sure value doesn't change between here and the next line...
			self.cc.frame_num_uplink[frame_idx] = frame_number #we could check to make sure that if it isn't already frame_num then 0 other nodes have reported...

			if mynum_reported == self.cc.num_bs_ant:
				self.cc.num_reported_uplink[frame_idx] = 0
				#grow the hdf5 file, if necessary
				if frame_number >= self.cc.UplinkData.shape[0] - 2:
					num_frames_grow = 2*self.cc._num_frames
					self.cc.UplinkData.resize(frame_number + num_frames_grow, axis=0)			
				self.cc.UplinkData[frame_number,:,:,:] = self.cc.UplinkData_buffer[frame_idx,:,:,:]
				self.cc.max_uplink_frame_complete = frame_number

			
		def connection_made(self, transport):
			self.logger.info('ArgosCC UplinkData server started... %s:%s' % transport.get_extra_info('sockname') )
			print('ArgosCC UplinkData server started... %s:%s' % transport.get_extra_info('sockname') )
			self.transport = transport
	
		def connection_refused(self, exc):
			self.logger.info('UplinkData Connection refused: %s' % str(exc))	

		def connection_lost(self, exc):
			self.logger.info('UplinkData Connection lost: %s' % str(exc) )		
	
		def loop_in_thread(self):
			"""A thread calls this to start the event loop to listen to datagrams on the socket."""
			asyncio.set_event_loop(self.loop)
			#self.loop.set_debug(True) #useful for debugging
			self.loop.run_forever()			
			
			
	class CSIReceiver:
		"""The CSIReceiver handles all packets sent by the WarpNodes that contain the pilot IQ samples -- it simply parses them out and saves them in the HDF5 file."""
		def __init__(self, cc, host_addr='0.0.0.0', port=55104):
			self._host_addr = host_addr
			self._host_port = port
			self.cc = cc
			self.logger = self.cc.logger
			self.loop = asyncio.new_event_loop()
			#self.loop.set_debug(True)
			self.loop.run_until_complete(self.loop.create_datagram_endpoint(lambda: self, local_addr=(self._host_addr,self._host_port) ) )
			self.socket = self.transport.get_extra_info('socket')

			sock_buf_size = int(1024*1024*64*2) #macs complain about a float...
			try:
				self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, sock_buf_size)  #this is critical for not dropping packets! #Note! In linux you have to add "net.core.rmem_max = 67108864" to /etc/sysctl.conf then do a sysctl -p to load the changes
			except:
				print("Warning!  There was a problem setting the receive socket buffer size.  This happens on Macs.  Be wary of dropped packets!  It should be %d, but is actually %d" % (sock_buf_size, self.socket.getsockopt(socket.SOL_SOCKET,socket.SO_RCVBUF)))
				self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 1024*1024*6*2) #todo check what actually fails on macs

			self.thread_loop = threading.Thread(target=self.loop_in_thread)
			self.thread_loop.start()	


		#note: windows firewall will block this by default.
		def datagram_received(self,data,addr):
			""" 
			This is what handles all pilot packets automatically sent by the nodes (e.g. CSI, but not any command responses, which are handled by the WarpNode object.)
			
			Currently, if the 'set_cc_macip' command is sent to the bs_nodes they will start reporting CSI and RSSI of the pilots to the specified MAC/IP.
			The packet format is FrameNumber CSI_RADIO_1_TERM_1 CSI_RADIO_1_TERM_2 .. CSI_RADIO_1_TERM_K CSI_RADIO_2_TERM_1 CSI_RADIO_2_TERM_2 ... CSI_RADIO_4_TERM_K, then the same pattern with RSSI.
			All values are reported as 4 bytes.
			
			Additionally, mobile reporting of beamformed downlink strength can also be reported by sending the 'set_cc_macip' to the mobile nodes.
			This is still being tested.
			
			This function parses those packets and saves them in the circular buffers.
			Once all nodes have reported for a given frame, the frame is recorded to disk by copying them to the h5py file.
			
			Note that this was intially designed to be multithreaded to improve performance, but the Python GIL makes multithreading essentially worthless.
			TODO: In the future, the recording to disk should only happen in chunks, not every frame.
			"""
			warp_header_len = 12
			csi_start = warp_header_len + 6 #2 bytes pad at beginning + sizeof(int), for frame_number
			#self.logger.info("data received from %s" % addr[0])
			#we're only sending the transport header, not the command/response header #http://warpproject.org/trac/wiki/WARPLab/Reference/Architecture/WireFormat
			header = struct.unpack(">xx2H xB 3H", data[0:14])  #> means big endian.  H is 2 bytes, B is 1, I is 3, x is disregarded single pad byte.  capital is unsigned
			ID = header[0]
			nodeID = header[1]
			pkt_type = header[2]
			length = header[3]
			seq_num = header[4]
			flags = header[5] 
	
			frame_number = struct.unpack("I", data[csi_start-4:csi_start])[0]			
			frame_idx = frame_number % self.cc._num_frames  #index in to our cache arrays (circular buffers)

			#print("Packet from node %d received!  Frame number %d, length %d. " % (nodeID, frame_number, len(data) ) )
			#self.logger.info("Packet from node %x with cmd %d (cmdID %d, cmdGrp %d) and %d cmdArgs received!  Frame number %d, length %d. " % (nodeID, cmd, cmdID, cmdGrp, numArgs, frame_number, len(data) ) )
			#self.logger.info("Packet from node %d received!  Frame number %d (idx %d), length %d (%d reported). " % (nodeID, frame_number, frame_idx, len(data), length ) )
			
			#make sure that we're not overwriting anything in our buffer before it is saved.
			if frame_number > self.cc.max_frame_complete + self.cc._num_frames/4 or frame_number < self.cc.max_frame_complete - self.cc._num_frames/4: 
				print("Warning! Frame number %d from node %d too far away from most recent complete frame number %d.  Discarding packet!" % (frame_number, nodeID, self.cc.max_frame_complete) ) #hrm, well if we drop too many packets in a row, then the frame number will *always* be too high.  Perhaps we should have a counter or something and reset after a while, or just only look to see if it is too low.
				return
			
			#mobile node reported (experimental)
			if nodeID in self.cc.MOBID2idx:
				#self.cc.logger.info("Received %d bytes of data from mobile for frame %d!" % (len(data), frame_number) )
				mob_rssi = np.fromstring(data[csi_start:csi_start+self.cc.num_mob_rssi*4], dtype='u4')
							
				if self.cc.collect_down_rssi:
					#self.cc.DownRSSI_buffer[frame_idx, self.cc.MOBID2idx[nodeID], :] = mob_rssi
					self.cc.DownRSSI[frame_number, self.cc.MOBID2idx[nodeID], :] = mob_rssi  #save directly (bypass buffer)
				return
			
			#record first timestamp
			if self.cc.ts_buffer[frame_idx,0] == 0:  
				self.cc.ts_buffer[frame_idx,0] = np.int64(time.clock()*1e7) #does this even have good enough resolution?
			
			rssi_len = self.cc.num_mob_ant*self.cc._num_ant_per_bs*4
			if self.cc.sample_dump > 512:  #if we want more than 512 samples, we have to seperate them in to one packet per bs antenna
				csi_len = self.cc.sample_dump*4 #4bytes per sample
				bs_ant_idx = struct.unpack("I", data[csi_start:csi_start+4])[0]-1 #the second arg is the antenna idx for consistency it starts with 1, so we have to subtract 1
				csi_start += 4 #skip the antenna idx
				rssi_len = self.cc.num_mob_ant*4
			elif self.cc.sample_dump: #wideband
				csi_len = self.cc.sample_dump*self.cc._num_ant_per_bs*4 #4bytes per sample		
			else: #narrowband
				csi_len = self.cc.num_mob_ant*self.cc._num_ant_per_bs*4 #4bytes per sample
			
			if len(data) != csi_start+csi_len + rssi_len + 2 or (len(data)-warp_header_len) != length:  #2 bytes padding at beginning and end
				print("Warning! Data length mismatch received pack of length %d, should be %d.  WARP header reported %d, should be %d  Discarding packet." % (len(data), csi_start+csi_len + rssi_len + 2, length, (len(data)-warp_header_len) ) ) #self.cc.logger.warn
				return
			
			if nodeID not in self.cc.BSID2idx:
				self.cc.logger.warn("Warning! Received data from bogus nodeID %d.  Discarding packet." % nodeID)
				return
			
			CSI_raw = np.fromstring(data[csi_start:csi_start+csi_len], dtype='i2') #use '>i2' for big-endian
			RSSI = np.fromstring(data[csi_start+csi_len:csi_start+csi_len+rssi_len], dtype='u4')
						
			CSI_Q = CSI_raw[0::2]  #these are reversed because of the endian swap.
			CSI_I = CSI_raw[1::2]
			
			#store to the buffer
			if self.cc.sample_dump > 512: #this is a wideband trace with one packet per BS antenna
				self.cc.Pilot_Sample_buffer[frame_idx, self.cc.BSID2idx[nodeID]*self.cc._num_ant_per_bs+bs_ant_idx,:, 0] = CSI_I #CSI_I[mob*self.cc.sample_dump:(mob+1)*self.cc.sample_dump]
				self.cc.Pilot_Sample_buffer[frame_idx, self.cc.BSID2idx[nodeID]*self.cc._num_ant_per_bs+bs_ant_idx,:, 1] = CSI_Q #CSI_Q[mob*self.cc.sample_dump:(mob+1)*self.cc.sample_dump]
				self.cc.RSSI_buffer[frame_idx, :, self.cc.BSID2idx[nodeID]*self.cc._num_ant_per_bs+bs_ant_idx] =  RSSI[:] #& 0xEFFFFFFF #np.int32(r)
				
				#no longer necessary with new timer based AGC
				#if self.cc.iq_agc_en:
				#	mag = np.sqrt(np.square(CSI_I.astype('int32')) + np.square(CSI_Q.astype('int32')))  #todo save users separately...
				#	user_mag = mag.reshape((self.cc.num_mob_ant,self.cc.samples_per_user))
				#	self.cc.temp_user_mag=user_mag
				#	self.cc.temp_mag=mag
				#	self.cc.CSI_max_buffer[frame_idx, :, self.cc.BSID2idx[nodeID]*self.cc._num_ant_per_bs+bs_ant_idx] = user_mag[:,32:-32].mean(axis=1) #don't use boundary samples... #todo:check this.
					
			else: #one packet per node				
				for bs in range(self.cc._num_ant_per_bs):
					for	mob in range(self.cc.num_mob_ant):
						self.cc.RSSI_buffer[frame_idx, mob, self.cc.BSID2idx[nodeID]*self.cc._num_ant_per_bs+bs] =  RSSI[self.cc.num_mob_ant*bs+mob] #& 0xEFFFFFFF #np.int32(r)			
					if self.cc.sample_dump:
						ant_CSI_I = CSI_I[bs*self.cc.sample_dump:(bs+1)*self.cc.sample_dump]
						ant_CSI_Q = CSI_Q[bs*self.cc.sample_dump:(bs+1)*self.cc.sample_dump]
						self.cc.Pilot_Sample_buffer[frame_idx, self.cc.BSID2idx[nodeID]*self.cc._num_ant_per_bs+bs,:, 0] = ant_CSI_I #CSI_I[mob*self.cc.sample_dump:(mob+1)*self.cc.sample_dump]
						self.cc.Pilot_Sample_buffer[frame_idx, self.cc.BSID2idx[nodeID]*self.cc._num_ant_per_bs+bs,:, 1] = ant_CSI_Q #CSI_Q[mob*self.cc.sample_dump:(mob+1)*self.cc.sample_dump]
						
						#no longer necessary with new timer based AGC
						#if self.cc.iq_agc_en:
						#	mag = np.sqrt(np.square(ant_CSI_I.astype('int32')) + np.square(ant_CSI_Q.astype('int32')))						
						#	user_mag = mag.reshape((self.cc.num_mob_ant,self.cc.samples_per_user))		
						#	umag = user_mag[:,32:-32].mean(axis=1)
						#	for	mob in range(self.cc.num_mob_ant):	
						#		if umag[mob] > self.cc.CSI_max_buffer[frame_idx, mob, self.cc.BSID2idx[nodeID]]:
						#			self.cc.CSI_max_buffer[frame_idx, mob, self.cc.BSID2idx[nodeID]] = umag[mob] #don't use boundary samples... #todo:check this.
					
					else:
						for	mob in range(self.cc.num_mob_ant):
							self.cc.CSI_buffer[frame_idx, mob, self.cc.BSID2idx[nodeID]*self.cc._num_ant_per_bs+bs, 0] = CSI_I[mob]
							self.cc.CSI_buffer[frame_idx, mob, self.cc.BSID2idx[nodeID]*self.cc._num_ant_per_bs+bs, 1] = CSI_Q[mob] 
							mag = math.sqrt(CSI_I[mob]**2 + CSI_Q[mob]**2)
							if mag > self.cc.CSI_max_buffer[frame_idx, mob, self.cc.BSID2idx[nodeID]]:
								self.cc.CSI_max_buffer[frame_idx, mob, self.cc.BSID2idx[nodeID]] = mag #& 0xEFFFFFFF #np.int32(mag)				
			
			if self.cc.iq_agc_en or self.cc.rssi_agc_en:
				for mob in self.cc.mob_nodes:
					for ifc in range(self.cc._num_ant_per_mob):
						self.cc.mob_tx_gain_buffer[frame_idx, self.cc.MOBID2idx[mob.ID]*self.cc._num_ant_per_mob + ifc, :] = mob.tx_gain_setting[mob.RFS[ifc]][:]
						
			with self.cc.num_reported_lock: #we could create _num_frames locks to avoid excessive contention, but it shouldn't help that much if each frame is processed in realtime
				self.cc.num_reported[frame_idx] += 1  #notify that one more node reported in
				mynum_reported = self.cc.num_reported[frame_idx] #make sure value doesn't change between here and the next line...
			self.cc.frame_num[frame_idx] = frame_number #we could check to make sure that if it isn't already frame_num then 0 other nodes have reported...

			if (mynum_reported == self.cc.num_bs_nodes and self.cc.sample_dump <= 512) or (mynum_reported == self.cc.num_bs_nodes*self.cc._num_ant_per_bs and self.cc.sample_dump > 512):
				self.cc.num_reported[frame_idx] = 0  #this was in the next if statement... not sure why
				if frame_number > self.cc.max_frame_complete:
					self.cc.max_frame_complete = frame_number
				self.cc.process_complete_frame(frame_idx)

		def connection_made(self, transport):
			self.logger.info('ArgosCC server started... %s:%s' % transport.get_extra_info('sockname') )
			print('ArgosCC server started... %s:%s' % transport.get_extra_info('sockname') )
			self.transport = transport
	
		def connection_refused(self, exc):
			self.logger.info('Connection refused: %s' % str(exc))
	
	
		def connection_lost(self, exc):
			self.logger.info('Connection lost: %s' % str(exc) )		
	
		def loop_in_thread(self):
			"""A thread calls this to start the event loop to listen to datagrams on the socket."""
			asyncio.set_event_loop(self.loop)

			#self.loop.set_debug(True)
			self.loop.run_forever()			
			
	zeroforcing_enabled = False
	
	_num_frames = 256  #how many frames to cache -- if we have packets arrive later than this #, then everything is messed up. #this also sets the hdf5 chunk size.
	_num_cc = 15   # number of calibration coefficients 
	avg_calib = False   # Average Calibration
	_num_ant_per_bs = 4  #we could read this from each node, but that gets pretty ugly pretty quick
	_num_ant_per_mob = 1  #we could read this from each node, but that gets pretty ugly pretty quick
	
	#current_frame_number = 0 #keeps track of the current frame number (modulus _num_frames) #there is no frame number... it is always specified in the packet
	BSID2idx = {}
	MOBID2idx = {}

	IQ_max = (2**15) #math.sqrt(2)#we don't unscale the magnitude, so it can be this big.
	
	def __init__(self, nodesFile = './argos.ips', frame_length = .2e6, pilot_length = 64, pilot_freq = 5e6, frequency=2.4, channel=14, wurc_channel=563000, iq_agc_en=True, rssi_agc_en=True, agc_interval=.05, collect_down_rssi=False, both_mob_pilots = False, num_mob_rssi=None, loglevel=logging.CRITICAL, mob_tx_gain=[3,30], mob_rx_gain=[1,10], bs_tx_gain=[1,10], bs_rx_gain=[1,15], wideband=True, bandwidth=20, num_uplink_samples=0, uplink_sample_rate=1.0, PilotGuard=80, wurc_tx_lpf_mhz=20, wurc_rx_lpf_mhz=20, dont_init=False):
		"""Initialize an instance of the ArgosCC."""
		
		self.nodes, self.mnode, self.bs_nodes, self.mob_nodes, bswl_nodes, mobwl_nodes = argos_init(reconfig_nodes = 'auto', frame_length = frame_length, pilot_length = pilot_length, pilot_freq = pilot_freq, frequency = frequency, channel = channel, wurc_channel=wurc_channel, extra_pilots = 1, both_mob_pilots = both_mob_pilots, nodesFile = nodesFile, wideband_pilot=wideband, bandwidth=bandwidth, wurc_tx_lpf_mhz=wurc_tx_lpf_mhz, wurc_rx_lpf_mhz=wurc_rx_lpf_mhz, PilotGuard=PilotGuard, dont_init=dont_init) #, wurc_tx_lpf_mhz=bandwidth, wurc_rx_lpf_mhz=bandwidth

		# create logger
		self.logger = logging.getLogger('argosCC.py')
		self.logger.setLevel(loglevel)
		# create console handler and set level to debug
		ch = logging.FileHandler('./logs/debuginfo.log')
		ch.setLevel(logging.DEBUG)
		# create formatter
		formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
		# add formatter to ch
		ch.setFormatter(formatter)
		# add ch to logger
		self.logger.addHandler(ch)
		
		self.num_bs_nodes = len(self.bs_nodes)
		self.num_mob = len(self.mob_nodes)
	
		#self._num_ant_per_mob = mob_nodes[0].num_interfaces
		self._num_ant_per_mob = 2 if both_mob_pilots else 1
		self.num_mob_ant = self.num_mob*self._num_ant_per_mob + 1 #we could query every node to see how many they have, but lets assume they all have the same # for now # +1 since last measurement is always a noise estimate!
		self.num_bs_ant = self.num_bs_nodes*self._num_ant_per_bs
	
		#save settings:
		self.pilot_length = pilot_length
		self.frame_length = frame_length
		self.pilot_freq = pilot_freq
		self.channel = channel
		self.frequency = frequency
		self.wurc_channel = wurc_channel
		self.bs_tx_gain = bs_tx_gain
		self.bs_rx_gain = bs_rx_gain
		self.init_mob_tx_gain = mob_tx_gain
		self.mob_rx_gain = mob_rx_gain	
		self.iq_agc_en = iq_agc_en
		self.rssi_agc_en = rssi_agc_en
		self.agc_interval = agc_interval  #if the uplink AGC control takes longer than a frame then it could cause nasty oscillations and break.  To support shorter frame times we have to skip frames.
		self.collect_down_rssi = collect_down_rssi
		self.num_uplink_samples = num_uplink_samples
		self.uplink_sample_rate = uplink_sample_rate #0.25 #testing uplink sample rate
		if num_mob_rssi is not None:
			self.num_mob_rssi = num_mob_rssi
		
		self.nodes.barrier() #ridiculous bug if you don't do this -- essentially settings from the init may not be done, then we either read the wrong values, or the init commands overwrite ours
		
		self.csi_port = 55104
		self.uplink_port = 52526
		[self.bs_nodes[i].nodeCmd('set_cc_macip', self.num_mob_ant, port=self.csi_port) for i in range(self.num_bs_nodes)]	
		if self.num_uplink_samples:
			self.bs_nodes.nodeCmd('dump_up_buff', self.num_uplink_samples, self.uplink_port) #length,port
		self.num_reported_lock = threading.Lock() #for incrementing the num_reported.
		
			
		if hasattr(self.bs_nodes[0].baseband, 'UP_SAMPLE_RATE'):  #temporary hack to not error before we upgrade the non-WURC baseband
			self.bs_nodes.basebandCmd('up_sample_rate', int(self.uplink_sample_rate * 2**4) )
		else:
			print("Warning! Old BS baseband only supports uplink sample rate of 1.0")
		
		self.wideband = wideband
		self.sample_dump = self.mob_nodes[0].basebandCmd('PILOTEND').result()[0] - self.mob_nodes[0].basebandCmd('PILOTSTART').result()[0]		
		self.sample_dump = int(self.sample_dump*self.uplink_sample_rate) #apply the sample rate	
		self.samples_per_user = int(self.sample_dump/self.num_mob_ant)
		if self.wideband:	
			self.bs_nodes.nodeCmd('dump_pilot_buff', self.sample_dump)
			#self.iq_agc_en = False #can't do AGC right now with sample_dump  TODO: enable RSSI based AGC
				
		self.nodes_descr = ''
		for n in self.nodes:
			if n.nodetype & n.ARGOSBS_TYPE: #bitmask for WURC
				self.BSID2idx[n.ID] = len(self.BSID2idx)
			elif n.nodetype & n.ARGOSMOBILE_TYPE:
				self.MOBID2idx[n.ID] = len(self.MOBID2idx) #these have to be the right order!
			self.nodes_descr += str(n)

		#initialize the gains to their setting (regardless of whether AGC is enabled)
		[n.interfaceCmd('tx_gains', n.RFALL, *bs_tx_gain) for n in self.bs_nodes]
		[n.interfaceCmd('rx_gains', n.RFALL, *bs_rx_gain) for n in self.bs_nodes]
		[n.interfaceCmd('tx_gains', n.RFALL, *mob_tx_gain) for n in self.mob_nodes]
		[n.interfaceCmd('rx_gains', n.RFALL, *mob_rx_gain) for n in self.mob_nodes]
		
		if self.num_bs_nodes == 0  or self.num_mob == 0:
			print('Error! There must be at least 1 BS node and 1 mobile node to collect CSI.  We found %d BS nodes and %d mobile nodes.' % (self.num_bs_nodes, self.num_mob) )
			return

		#preallocate some arrays to store our stuff:
		#supposedly numpy is actually faster at this than builtin ways: http://stackoverflow.com/questions/5347108/how-to-preallocate-a-list-of-lists
		if self.wideband:	
			self.Pilot_Sample_buffer = np.empty((self._num_frames, self.num_bs_ant, self.sample_dump, 2), dtype=np.int16) #separate IQ #don't use complex since it's too big and takes time to convert.
			self.CSI_max_buffer = np.zeros((self._num_frames, self.num_mob_ant, self.num_bs_ant), dtype=np.uint32) #we're going to need this for doing uplink AGC, so let's compute it while we wait on other packets
		else:
			self.CSI_buffer = np.empty((self._num_frames, self.num_mob_ant, self.num_bs_ant, 2), dtype=np.int16) #separate IQ #don't use complex since it's too big and takes time to convert.
			self.CSI_max_buffer = np.zeros((self._num_frames, self.num_mob_ant, self.num_bs_nodes), dtype=np.uint32) #we're going to need this for doing uplink AGC, so let's compute it while we wait on other packets
		
		self.RSSI_buffer = np.empty((self._num_frames, self.num_mob_ant, self.num_bs_ant), dtype=np.uint32) 
		self.ts_buffer = np.zeros((self._num_frames,2), dtype=np.uint64) 
		if self.iq_agc_en or self.rssi_agc_en:
			self.mob_tx_gain_buffer = np.empty((self._num_frames, self.num_mob_ant-1, 2), dtype=np.uint8) #-1 since last one is noise		
			#self.target_dB = InterfaceM2829.target_dB_fromRxGain(*self.bs_rx_gain) #todo: fix this for non WURC
			self.target_dB = -self.bs_rx_gain[0] - 14  #-15 is about .5 
			self.target_IQ = .3 #1/2  #target average IQ.  .5 usually gives up to .8 max, so we want to be conservative.
		self.frame_num = np.empty((self._num_frames), dtype=np.int32) #frame number corresponding to buffer index
		if self.collect_down_rssi:
			self.DownRSSI_buffer = np.empty((self._num_frames, self.num_mob_ant-1, self.num_mob_rssi), dtype=np.uint32) 
		
		self.num_reported = np.zeros((self._num_frames), dtype=np.int16) # we have to init this here since the processing loop looks at it.

		if self.num_uplink_samples:
			self.UplinkData_buffer = np.empty((self._num_frames, self.num_bs_ant, self.num_uplink_samples, 2), dtype='int16')
			self.num_reported_uplink = np.zeros((self._num_frames), dtype=np.int16)
			self.frame_num_uplink = np.empty((self._num_frames), dtype=np.int32) #frame number corresponding to buffer index
		
		self.use_CLIB = False
		if self.sample_dump > 512 and 'Argos_HDF5_CLIB' in globals():
			self.use_CLIB = True
		#Right now the nodes are configured to send a single packet for all 4 antennas if it fits in one jumbo frame (to reduce overhead)
		#The C implementation *only* supports 1 packet per antenna (not per node)
		elif self.sample_dump <= 512:
			print("Warning! For legacy reasons, C acceleration requires num_uplink_samples to be greater than 512.  Using slower Python implementation.")

		if frame_length < .2e6 and not self.use_CLIB:
			print("Warning!  Python implementation may not support this short of frame, %f s" % (.2e6/20e6) )
		
		self.stop_running = False #flag to stop threads
		
		max_ant = (self.mnode._mtu - 24) // (self.samples_per_user * 4)
		if max_ant < self.num_mob_ant:
			raise ValueError("The current settings and MTU (%d) can only support up to %d users, but you have configured %d.  Please double checking your settings.  Have you enabled jumbo frames?" % (self.mnode._mtu, max_ant, self.num_mob_ant))

	'''		
	def calibrate(self, avg_calib=False):
		"""Implicit reciprocal calibration."""
		acc = ascc = np.zeros((self.num_bs_ant,self.num_bs_ant))
		if avg_calib:
			for i in range(_num_cc):
				if i > 0:
					acc = (acc+cc/(i))  		# averages calibration coefficients
					ascc =(ascc+scc/(i))       # averages scaled calibration coefficients
				cc, scc, ce = argos_calibrate(bs_nodes, mnode)
			self.calibration = self.h5file.create_dataset("Average Calibration Cofficients", (1, self.num_bs_ant, self.num_bs_ant), dtype='complex64', maxshape=(None, self.num_bs_ant, self.num_bs_ant), chunks=(1, self.num_bs_ant, self.num_bs_ant))
		else:
			self.calibration = self.h5file.create_dataset("Calibration Cofficients", (self._num_cc, self.num_bs_ant, self.num_bs_ant), dtype='complex64', maxshape=(None, self.num_bs_ant, self.num_bs_ant), chunks=(self._num_cc, self.num_bs_ant, self.num_bs_ant))		
	'''	
	
	def init_h5file(self, filename=None, init_datasets=True):
		"""
		Initialize a new h5file with the correct data structure and attributes.
		We do this outside of __init__ so that multiple files can be collected without reinitializing the argosCC.
		"""
		#there are *lots* of options on how to save the channel estimates and other info.
		#The other two good options would probably be pytables or a direct save from numpy
		#https://www.pytables.org/moin/FAQ#HowdoesPyTablescomparewiththeh5pyproject.3F
		#http://docs.scipy.org/doc/numpy/reference/routines.io.html
		#Really all we want to do is to save the data as we go along, so we really don't need anything fancy, we just want it as fast as possible.
		#PyTables is probably too much, and numpy save is not very compatible, so let's stick with h5py for now.  If it's too slow, it may be faster to just have numpy write directly to a file.

		#note: Matlab can read these files directly via something like: h5disp('ArgosCSI-80x3-2015-05-15-00-37-47.hdf5') or h5read('ArgosCSI-80x3-2015-05-15-00-37-47.hdf5') 
		
		#set the hdf5 filename
		if filename is None:
			if not os.path.exists('logs'):
				os.makedirs('logs')
			self.filename = ("logs/ArgosCSI-%dx%d-" % (self.num_bs_ant, self.num_mob_ant-1) )+ time.strftime("%Y-%m-%d-%H-%M-%S.hdf5")
		else:
			self.filename = filename		
		
		#create hdf5 file and setup the structure and attributes.
		self.h5file = h5py.File(self.filename, "w")
		self.logger.debug(self.h5file)
		if init_datasets:
			self.init_datasets()
		self.h5file.attrs['nodes_description'] = self.nodes_descr.encode('utf8')
		self.h5file.attrs['pilot_length'] = self.pilot_length  
		self.h5file.attrs['frame_length'] = self.frame_length  #time in microseconds
		self.h5file.attrs['pilot_frequency'] = self.pilot_freq  
		self.h5file.attrs['channel'] = self.channel 
		self.h5file.attrs['wurc_channel'] = self.wurc_channel  #if we were consistent this wouldn't be wurc_....
		self.h5file.attrs['num_bs_nodes'] = self.num_bs_nodes  #probably redundant
		self.h5file.attrs['num_bs_ant'] = self.num_bs_ant  #probably redundant
		self.h5file.attrs['num_mob'] = self.num_mob  
		self.h5file.attrs['num_mob_ant'] = self.num_mob_ant - 1  #-1 since the last one is noise.
		self.h5file.attrs['start_time'] = time.time()
		self.h5file.attrs['start_clock'] = time.clock()
		self.h5file.attrs['bs_tx_gain'] = self.bs_tx_gain
		self.h5file.attrs['bs_rx_gain'] = self.bs_rx_gain
		self.h5file.attrs['BSID2idx'] = str(self.BSID2idx)
		
		cur_mob_tx_gain = []
		for mob in self.mob_nodes:
			for ifc in range(self._num_ant_per_mob):
				cur_mob_tx_gain.append(mob.tx_gain_setting[mob.RFS[ifc]])
		self.h5file.attrs['mob_tx_gain'] = cur_mob_tx_gain #self.init_mob_tx_gain #[1,15] #todo: read these or set these
		self.h5file.attrs['mob_rx_gain'] = self.mob_rx_gain #[1,15] #todo: read these or set these  #hrm, if agc isn't enabled then we should either not save gains for each frame, or set them to the static value...
		if self.iq_agc_en or self.rssi_agc_en:
			self.h5file.attrs['mob_tx_gain'] = [-1,-1] #if AGC is enabled the mobile gains will be dynamic.		
			self.h5file.attrs['agc_interval'] = self.agc_interval 
		self.h5file.attrs['iq_agc_en'] = self.iq_agc_en 
		self.h5file.attrs['rssi_agc_en'] = self.rssi_agc_en 
		#self.h5file.attrs['num_subcarriers'] = self.num_sc
		if self.collect_down_rssi:
			self.h5file.attrs['mobile_rssi_avg_length'] = 32 #todo: read these or set these
			self.h5file.attrs['num_mob_rssi'] = self.num_mob_rssi
		self.h5file.attrs['wideband'] = self.wideband
		if self.wideband:
			self.h5file.attrs['samples_per_user'] = self.samples_per_user
			self.h5file.attrs['sample_dump'] = self.sample_dump
		self.h5file.attrs['num_uplink_samples'] = self.num_uplink_samples
		self.h5file.attrs['uplink_sample_rate'] = self.uplink_sample_rate
			
	def	init_datasets(self, group=None):
		"""Initialize a dataset in the current hdf5 file.  
		We do this separately so we can collect multiple traces in a single hdf5 file, e.g. for AoA calibration.
		"""
		self.h5group = "" if group is None else group
		group = self.h5file if group is None else self.h5file.create_group(group)
		if self.wideband:
			#TODO: undo length hack for c code *1000
			self.Pilot_Samples = group.create_dataset("Pilot_Samples", (self._num_frames, self.num_bs_ant, self.sample_dump, 2), dtype='int16', maxshape=(None, self.num_bs_ant, self.sample_dump,2), chunks=(self._num_frames, self.num_bs_ant, self.sample_dump, 2))
			group['Pilot_Samples'].attrs['column_names'] = [a.encode('utf8') for a in ['Frame','BS_Ant','Samples','IQ'] ]		
		else:
			self.CSI = group.create_dataset("CSI", (self._num_frames, self.num_mob_ant, self.num_bs_ant, 2), dtype='int16', maxshape=(None, self.num_mob_ant, self.num_bs_ant,2), chunks=(self._num_frames, self.num_mob_ant, self.num_bs_ant,2))
			group['CSI'].attrs['column_names'] = [a.encode('utf8') for a in ['Frame','Mobile','BS_Ant','IQ'] ]
		self.RSSI = group.create_dataset("RSSI", (self._num_frames, self.num_mob_ant, self.num_bs_ant), dtype='uint32', maxshape=(None, self.num_mob_ant, self.num_bs_ant), chunks=(self._num_frames, self.num_mob_ant, self.num_bs_ant))
		group['RSSI'].attrs['column_names'] =  [a.encode('utf8') for a in ['Frame','Mobile','BS_Ant']] #['Frame','Mobile','BS_Ant'] .encode('utf8')
		self.timestamp = group.create_dataset("FrameCompleteTime", (self._num_frames,2), dtype='uint64', maxshape=(None,2), chunks=(self._num_frames,2))
		#if self.iq_agc_en or self.rssi_agc_en:
		#	self.mob_tx_gain = group.create_dataset("mob_tx_gain", (self._num_frames, self.num_mob_ant-1, 2), dtype='uint8', maxshape=(None, self.num_mob_ant-1, 2), chunks=(self._num_frames, self.num_mob_ant-1, 2))
		if self.collect_down_rssi:
			self.DownRSSI = group.create_dataset("DownRSSI", (self._num_frames, (self.num_mob_ant-1), self.num_mob_rssi), dtype='uint32', maxshape=(None, (self.num_mob_ant-1), self.num_mob_rssi), chunks=(self._num_frames, (self.num_mob_ant-1), self.num_mob_rssi))
		if self.num_uplink_samples:
			self.UplinkData = group.create_dataset("UplinkData", (self._num_frames, self.num_bs_ant, self.num_uplink_samples, 2), dtype='int16', maxshape=(None, self.num_bs_ant, self.num_uplink_samples, 2), chunks=(self._num_frames, self.num_bs_ant, self.num_uplink_samples, 2))
		

	def process_complete_frame(self, frame_idx):	
		"""
		Process a complete frame -- then save the frames to the hdf5 file, in chunks.
		This could be done in the CSIReceiver.
		"""

		ts_start = time.clock()
		frame_number = self.frame_num[frame_idx]
		self.ts_buffer[frame_idx,1] = np.int64(time.clock()*1e7) #does this even have good enough resolution?
		self.frame_num[frame_idx] = frame_number
		#self.logger.info('Frame %d with idx %d had all %d nodes report!' % (frame_number, frame_idx, self.num_bs_nodes) )
		#print('Frame %d with idx %d had all %d nodes report!' % (frame_number, frame_idx, self.num_bs_nodes) )
					
		
		if frame_number != self.max_frame_complete: #make sure we don't use old frames to set agc.	(this can only happen if we are receiving a frame very, very, late)		
			self.logger.warn("Warning! Received complete frame %d *after* frame %d." % (frame_number, self.max_frame_complete) ) #could need to increase buffer size...		
		''' AGC now done with timer
		else:
			if (self.iq_agc_en or self.rssi_agc_en) and (frame_number - self.last_agc_frame >= self.agc_interval):
				self.last_agc_frame = frame_number
				maxCSI, maxRSSI = self.get_max_amp(frame_idx)
				#self.run_agc(frame_idx)
				self.run_agc(frame_number, maxCSI, maxRSSI)
			if self.zeroforcing_enabled:
				#bws = np.linalg.pinv( np.transpose(self.CSI_buffer[frame_idx, :-1, :, 0] + self.CSI_buffer[frame_idx, :-1, :, 1]*1j) ) #index to -1 since last value is noise estimate.
				bws = np.linalg.pinv( (self.CSI_buffer[frame_idx, :-1, :, 0] + self.CSI_buffer[frame_idx, :-1, :, 1]*1j).astype(np.complex64) ) #index to -1 since last value is noise estimate.  #this casts to a complex128, when really we should only need a complex64 (or really less, but that's the smallest complex numpy supports)
				bws /= np.linalg.norm(bws,ord=np.inf)/self.IQ_max  #normalize.  #we could/should normalize per use as well...
		'''
		self.save_frame(frame_idx)

	def save_frame(self, frame_idx):
		"""
		Save a complete frame (all nodes have reported) to the hdf5 file.
		Extend the hdf5 file as needed.
		"""
		frame_number = self.frame_num[frame_idx]
		cur_len = self.Pilot_Samples.shape[0] if self.sample_dump else self.CSI.shape[0]
		#TODO: make the saving process a separate thread and write in batches!  Also save incomplete frames!
		#turns out because of the GIL this doesn't help much.		
		if frame_number >= cur_len - 1:
		#if frame_number == cur_len - self._num_frames//2: #only resize once, since we're multithreaded now
			num_frames_grow = 2*self._num_frames
			if self.sample_dump:
				self.Pilot_Samples.resize(frame_number + num_frames_grow, axis=0)
			else:
				self.CSI.resize(frame_number + num_frames_grow, axis=0)
			self.RSSI.resize(frame_number + num_frames_grow, axis=0)
			if self.iq_agc_en or self.rssi_agc_en:
				self.mob_tx_gain.resize(frame_number + num_frames_grow, axis=0)
			self.timestamp.resize(frame_number + num_frames_grow, axis=0)
			if self.collect_down_rssi:
				self.DownRSSI.resize(frame_number + num_frames_grow, axis=0)
			#let the uplink samples resize itself
			#if self.num_uplink_samples:
			#	self.UplinkData.resize(frame_number + num_frames_grow, axis=0)
		#Write the buffers to disk
		if self.sample_dump:
			self.Pilot_Samples[frame_number,:,:,:] = self.Pilot_Sample_buffer[frame_idx,:,:,:]			
		else:
			self.CSI[frame_number, :,:,:] = self.CSI_buffer[frame_idx, :,:,:]
		#self.logger.debug(self.CSI[frame_number, :,:,:])
		self.RSSI[frame_number, :,:] = self.RSSI_buffer[frame_idx, :,:]
		if self.iq_agc_en or self.rssi_agc_en:
			self.mob_tx_gain[frame_number,:,:] = self.mob_tx_gain_buffer[frame_idx,:,:]
		self.timestamp[frame_number,:] = self.ts_buffer[frame_idx,:]
		self.ts_buffer[frame_idx,:] = 0 #we check if 0 when setting the min ts.
		#self.h5file.flush() #save to disk
		
		self.max_frame_complete = frame_number
		self.num_reported[frame_idx] = 0
		#if not self.sample_dump:
		self.CSI_max_buffer[frame_idx, :, :] = 0	
		ts_end = time.clock()
		
		#self.logger.info("start: %f end: %f  time: %f" % (ts_start, ts_end, ts_end-ts_start) )
		#self.logger.info('total received for frame %d: %d' % (frame_number, self.num_reported[frame_idx]) )


	def start_workers(self):
		"""Start Python threads to collect CSI and Uplink packets (if enabled)."""
		#note that due to the GIL this multithreading actually slows things down.
		self.num_workers = 1 #12 #multiprocessing.cpu_count() // 2 - 1 # leave a thread for the main worker
		#self.num_workers = self.num_bs_nodes if self.num_workers > self.num_bs_nodes else self.num_workers
		#nodes_per_port = self.num_bs_nodes // self.num_workers
		#[self.bs_nodes[i].nodeCmd('set_cc_macip', self.num_mob_ant, port=self._port_start+i//nodes_per_port) for i in range(self.num_bs_nodes)]		
		#[self.bs_nodes[i].nodeCmd('set_cc_macip', self.num_mob_ant, port=self.csi_port) for i in range(self.num_bs_nodes)]		
		
		self.csi_workers = []
		for i in range(self.num_workers):
			self.csi_workers.append(self.CSIReceiver(self,port=self.csi_port+i))
		
		self.uplink_workers = []
		if self.num_uplink_samples:
			self.num_reported_uplink_lock = threading.Lock() #for incrementing the num_reported_uplink.
			self.bs_nodes.nodeCmd('dump_up_buff', self.num_uplink_samples, self.uplink_port) #length,port
			self.uplink_workers.append(self.UplinkDataReceiver(self,port=self.uplink_port))	
		

	def stop_workers(self):
		"""Stop all python threads for collecting CSI and uplink packets."""
		[w.transport.close() for w in self.csi_workers+self.uplink_workers]
		[w.loop.call_soon_threadsafe(w.loop.stop) for w in self.csi_workers+self.uplink_workers] #self.loop.stop() is not threadsafe.
		for w in self.csi_workers+self.uplink_workers:
			while w.loop.is_running():
				time.sleep(.05)
			w.loop.close()
			w.thread_loop.join()
		self.csi_workers = []
		self.uplink_workers =[]
		
	def get_max_amp(self,frame_idx=None):
		"""Return the max IQ and RSSI for each user across all antennas for a given frame, or the most recently complete frame if not provided.  Used for AGC."""
		frame_idx = frame_idx if frame_idx is not None else self.max_frame_complete % self.cc._num_frames
		maxIQ = []
		maxRSSI = []
		for mob in self.mob_nodes:
			for ifc in range(self._num_ant_per_mob):
				maxRSSI.append(self.RSSI_buffer[frame_idx, self.MOBID2idx[mob.ID]*self._num_ant_per_mob + ifc, :].max())  #todo, this 2^6 is based on the pilot length -- we should divide based on the pilot length.
				maxIQ.append(self.CSI_max_buffer[frame_idx, self.MOBID2idx[mob.ID]*self._num_ant_per_mob + ifc, : ].max() / self.IQ_max )  #this should work for either
		return maxIQ, maxRSSI


	def start_cworkers(self, ignore_incomplete_frame=False):
		"""
		Use C threads to write to HDF5 file
		Currently this calls ./ArgosCC.so, which is built by using the "make" in C/hd5dumper/ then installed via 'make install' or 'make install-svn'
		This requires the h5cc to be compiled an installed, e.g. via C/install_hdf5.sh
		This uses python/argos_hdf5_clib.py to call functions in main.c
		The main data writing loop is in AssembleBuffer.c
		"""		
		mCLib = Argos_HDF5_CLIB()
		mCLib.configCC(self.samples_per_user, self.csi_port, self.uplink_port, ignore_incomplete_frame)  
		mCLib.startCC(self.filename,self.h5group,self.num_uplink_samples)

		for m,d in enumerate(self.BSID2idx):  #make sure the python mapping corresponds to the C mapping
    			mCLib.setIdMap(d,m) #This function must be called after 'startCC' and before 'ENSYNCOUT is set to 1'. 	      

		self.mCLib = mCLib

		
	def collectCSI(self, duration=1, filename=None, calibrateAoA=False, calibration_duration=2, delay=0):
		"""
		Primary external function to collect CSI.
		Initializes the HDF5 file and data sets, sets up packet receiver threads, configures WARP nodes, and collects CSI, then does cleanup.
		It optionally collects AoA calibration data to sub-groups in the HDF5 file.
		"""
		time.sleep(delay)
		self.bs_nodes.basebandCmd('RADIOEN',0xFFFFFFFF)	#make sure the radios are on
		self.init_h5file(filename=filename, init_datasets=not calibrateAoA)
		if calibrateAoA:
			self.calibrateAoA(duration=calibration_duration)
			self.init_datasets()
		self.startCollection(duration)	
		self.stop()
		self.bs_nodes.basebandCmd('RADIOEN',0) #the WURC radios get super hot, so turn them off.

	def calibrateAoA(self, duration=1):
		"""Automate collection of traces at specific angles to help with AoA calibration.  Sets up sub-groups for front, left, and right and takes small traces at each location."""
		input("Put Mobile 0 in front of the array then press Enter to continue...")
		self.init_datasets('AoA_Front')
		self.startCollection(duration)
		input("Put Mobile 0 to the left of the array (while looking at it) then press Enter to continue...")
		self.init_datasets('AoA_Left')
		self.startCollection(duration)
		input("Put Mobile 0 to the right of the array (while looking at it) then press Enter to continue...")
		self.init_datasets('AoA_Right')
		self.startCollection(duration)
		input("Put Mobile 0 above the array then press Enter to continue...")
		self.init_datasets('AoA_Top')
		self.startCollection(duration)
		input("Press Enter to collect traces...")
		
	def startCollection(self, duration=1):
		"""Main helper function for collecting a single CSI trace (e.g., AoA or normal collection calls this.)."""
		if self.use_CLIB:
			self.h5file.flush()
			self.h5file.close() #we have to close it for the clib to be able to use it
			self.start_cworkers()			
		else:
			self.start_workers()
			self.num_reported = np.zeros((self._num_frames), dtype=np.int16) #count number of nodes that have reported for each frame (once it reaches num_bs_nodes the frame is ready to be processed)	
			if self.num_uplink_samples:
				self.num_reported_uplink = np.zeros((self._num_frames), dtype=np.int16)	
				for u in self.uplink_workers:
					u.counter = 0
			self.max_frame_complete = -1 #this keeps track of the max complete packet since packets can arrive out of order, let's make sure we always use the latest (this isn't necessary, but probably good practice)
			self.max_uplink_frame_complete = -1		
			#self.last_agc_frame = -self.agc_interval  # we do this with a timer now.
			
		if self.iq_agc_en or self.rssi_agc_en: #initialize gain settings
			self.gain_settings = [{} for i in range(self.num_mob_ant-1)] #dicts to store gain changes.
			mobid = -1
			for mob in self.mob_nodes:
				for ifc in range(self._num_ant_per_mob):
					mobid = mobid + 1
					self.gain_settings[mobid][0] = mob.tx_gain_setting[mob.RFS[ifc]][0]  #todo, don't store as list #initialize the gain settings.  This is a bit tedious now :/
					#print(mobid,mob.tx_gain_setting[mob.RFS[ifc]][0])		
					
		self.mob_nodes.basebandCmd('RESETFRAMECOUNT',1)
		self.mob_nodes.barrier()
		self.mob_nodes.basebandCmd('RESETFRAMECOUNT',0)
		self.bs_nodes.basebandCmd('ENSYNCOUT',0)
		
		self.nodes.barrier()
		[n.basebandCmd('CONTINUOUSFRAMES', 1) for n in self.mob_nodes]
		for n in self.bs_nodes:
			if n.ID != self.mnode.ID:			
				n.basebandCmd('ENSYNCOUT', 1) #You *must* enable ensyncout on all the bs boards (not just do a single sync) otherwise the frame numbers are reset.
		self.bs_nodes.barrier()
		self.mnode.basebandCmd('ENSYNCOUT', 1)
		self.mnode.barrier()

		if self.iq_agc_en or self.rssi_agc_en:
			for t in range(0,int(duration/self.agc_interval)):
				if self.use_CLIB:
					this_frame = self.mCLib.readFrame()
					maxIQ = self.mCLib.readMaxCSI()
					maxRSSI = self.mCLib.readMaxRSSI()
				else:
					this_frame = self.max_frame_complete % self._num_frames
					maxIQ,maxRSSI = self.get_max_amp(this_frame)
				
				#print('Frame %d, RSSI: %s'  % (this_frame, str(maxRSSI)))
				self.run_agc(this_frame,maxIQ,maxRSSI)
				
				time.sleep(self.agc_interval)
		else:
			time.sleep(duration)
		
		self.mnode.basebandCmd('ENSYNCOUT', 0)
		self.mnode.barrier()
		[n.basebandCmd('ENSYNCOUT', 0) for n in self.bs_nodes]
		
		[n.basebandCmd('CONTINUOUSFRAMES', 0) for n in self.mob_nodes]
		self.nodes.barrier()
		print("BS nodes received %s syncs." % str(self.bs_nodes.basebandCmd('FRAMENUMBER').result()) )
		print("Mobile nodes received %s syncs." % str(self.mob_nodes.basebandCmd('CORRELATOR_1_COUNT').result()) )  #TODO save this, and total number of frames to h5py file!
		fns = [n.basebandCmd('FRAMENUMBER').result()[0] for n in self.bs_nodes]
		if self.use_CLIB:
			self.max_frame_complete = self.mCLib.readFrame()
		print('Collection Done %d frames out of %d.  Waiting for processing to finish...' % (self.max_frame_complete+1, fns[0]))

		if not all([f == fns[0] for f in fns]):
			print('Error!  Not all the BS nodes have the same frame number! %s' % str(fns))
		

		if self.use_CLIB:
			time.sleep(2)
			self.mCLib.pauseCC()
			self.max_frame_complete = self.mCLib.readFrame()
			self.h5file = h5py.File(self.filename,'r+') #reopen hdf5 so we can clean it and save gains
		else:
			i = 0
			while (self.max_frame_complete+1 != fns[0]) and i < 40:			
				time.sleep(.2) #there's a backlog of rx packets still being processed.
				i += 1
			if self.num_uplink_samples:
				i = 0
				while (self.max_uplink_frame_complete+1 != fns[0]) and i < 20:
					#print(self.max_uplink_frame_complete)
					time.sleep(.3) #there's a backlog of rx packets still being processed.
					i += 1			
			self.stop_workers()
		self.clean_datasets()
		self.save_agc_gains() #to save space and maintain compatibility with C, we save AGC gain steps in a buffer and only add them to the HDF5 file after the trace is complete.
		print('Collected %d frames (out of %d)' % (self.max_frame_complete+1, fns[0]) )


	def run_agc(self, frame, maxCSI=None, maxRSSI=None):
		"""
		Given max CSI and/or RSSI for the most recently complete frame, this function sets the appropriate gains on the mobile.
		Note that this is setup with these input arguments so that it is compatible with the CLIB (which requires python to set node gains).
		"""
		#self.logger.info("-------------Doing uplink AGC---------------\n\n\n")
		#todo: check if pilot simply wasn't sent (e.g. if all radios below a threshold, and one of the last couple pilots were strong)! (then don't do agc)
		#todo: enable both iq and rssi mag
		mobid = -1
		for mob in self.mob_nodes:
			for ifc in range(self._num_ant_per_mob):
				mobid = mobid + 1
				cur_tx_gain_db = mob.tx_gain_setting[mob.RFS[ifc]] 
				cur_tx_gain_db = cur_tx_gain_db[0] if len(cur_tx_gain_db) == 1 else mob.IFCS[mob.RFS[ifc]].txgain2db(*cur_tx_gain)				
				new_tx_gain_db = cur_tx_gain_db
				#out_of_range = maxCSI[mobid] > 1 or maxCSI[mobid] < .1				
				#todo: check if we're clipping and RSSI enabled, then do only RSSI, otherwise IQ only (if enabled)				
				if self.rssi_agc_en and mob.RFS[ifc] & mob.RFALL_M2829: #Only maxim currently reports rssi.
					#1. find current max RSSI, which we use to adjust gains.  2. check target RSSI based on bs rx gain settings. 3. find dB difference between target and max RSSIs. 4. Convert dB gain to new mobile tx gain, and 5. send new gains
					#maxRSSI = self.RSSI_buffer[frame_idx, self.MOBID2idx[mob.ID]*self._num_ant_per_mob + ifc, :].max() / 2**(np.log2(self.pilot_length))  #todo, this 2^6 is based on the pilot length -- we should divide based on the pilot length.
					dbDiff = self.target_dB - InterfaceM2829.rssi2dB(maxRSSI[mobid] / 2**(np.log2(self.pilot_length)),InterfaceM2829.db2rxgain(self.bs_rx_gain[0])[0]) #self.bs_rx_gain[0])
					new_tx_gain_db = cur_tx_gain_db + dbDiff
					#print(cur_tx_gain_db, new_tx_gain_db, self.target_dB,maxRSSI[mobid]/64,InterfaceM2829.rssi2dB(maxRSSI[mobid]/64,InterfaceM2829.db2rxgain(self.bs_rx_gain[0])[0]))

				max_mag = maxCSI[mobid] 
				if self.iq_agc_en and ((max_mag < .85 and max_mag > .15) or not self.rssi_agc_en):	
				#if self.iq_agc_en and (not out_of_range or not self.rssi_agc_en):
				#if self.iq_agc_en:
					new_tx_gain_db = cur_tx_gain_db #reset it, in case RSSI AGC already adjusted it.
					#self.logger.info ('Frame %d: max IQ mag of mobile %d, idx %d, is %f, with tx_gain %d %d     IQ_max: %x' % (frame_number, mob.ID, self.MOBID2idx[mob.ID], max_mag, cur_tx_gain[0], cur_tx_gain[1], self.IQ_max) )
					#print('Frame %d: max IQ mag of mobile %d, idx %d, is %f, with tx_gain %d    IQ_max: %x' % (frame_number, mob.ID, self.MOBID2idx[mob.ID], max_mag, cur_tx_gain[0], self.IQ_max) )
					
					if max_mag > 2 * self.target_IQ:
						new_tx_gain_db -= 10
					elif max_mag > 1.6 * self.target_IQ:
						new_tx_gain_db -= 5	
					elif max_mag > 1.4 * self.target_IQ:
						new_tx_gain_db -= 3	
					elif max_mag > 1.2 * self.target_IQ:
						new_tx_gain_db -= 1.5					
					elif max_mag < .01 * self.target_IQ:
						new_tx_gain_db += 18				
					elif max_mag < .05 * self.target_IQ:
						new_tx_gain_db += 10					
					elif max_mag < .1 * self.target_IQ:
						new_tx_gain_db += 8					
					elif max_mag < .3 * self.target_IQ:
						new_tx_gain_db += 6				
					elif max_mag < .5 * self.target_IQ:
						new_tx_gain_db += 3				
					elif max_mag < .85 * self.target_IQ:
						new_tx_gain_db += 1.5	
						
				#check to make sure we are not outside the interface limits
				new_tx_gain_db = new_tx_gain_db if new_tx_gain_db <= mob.IFCS[mob.RFS[ifc]].tx_gain_max else mob.IFCS[mob.RFS[ifc]].tx_gain_max 
				new_tx_gain_db = new_tx_gain_db if new_tx_gain_db >= mob.IFCS[mob.RFS[ifc]].tx_gain_min else mob.IFCS[mob.RFS[ifc]].tx_gain_min 
										
				#if new_tx_gain_db != cur_tx_gain_db:
				if abs(new_tx_gain_db - cur_tx_gain_db) >= 1.5: #hysteresis
					mob.interfaceCmd('tx_gains', mob.RFS[ifc], new_tx_gain_db)					
					self.gain_settings[mobid][frame] = new_tx_gain_db #dicts to store gain changes.
					#print('Frame %d: Mobile %d, idx %d, tx gains old_db = %f new_db = %f' % (frame, mob.ID, self.MOBID2idx[mob.ID], cur_tx_gain_db,new_tx_gain_db ) )
				
						
		
	def save_agc_gains(self):
		"""Gain settings are appended to a dict, then we count how many entries there are and create an HDF5 dataset with the values."""
		if self.iq_agc_en or self.rssi_agc_en: # save the gain settings
			if self.h5file is None or not self.h5file.id: #if it's closed, e.g. for the C version, open it
				self.h5file = h5py.File(self.filename,'r+')
			group = self.h5file if len(self.h5group) == 0 else self.h5file[self.h5group]
			max_frame = group['Pilot_Samples'].shape[0] #todo check if narrowband...
			num_gain = np.max([len(x) for x in self.gain_settings])+1 #+1 since we append one for the last frame
			self.mob_tx_gain = group.create_dataset("mob_tx_gain", (num_gain, self.num_mob_ant-1, 2), dtype='int32') #max_num_gains, mobile antenna, (frame,gain setting in DB)
			for m,d in enumerate(self.gain_settings):
				i = 0
				for k,v in sorted(d.items()): #sort here, since ordereddicts are super slow.	
					self.mob_tx_gain[i,m,:] = [k,v]
					i += 1
				#print(i,max_frame,v)
				self.mob_tx_gain[i,m,:] = [max_frame,v] #makes plotting cleaner/easier/clearer
	
	def clean_datasets(self):
		"""Resize the h5py datasets so that we don't have all the zeros at the end"""
		if self.h5file is not None and self.h5file.id: #it seems the easy way to check if the h5file is open is just if file.id
			group = self.h5file if len(self.h5group) == 0 else self.h5file[self.h5group]
			total_frames = self.max_frame_complete+1 #since we start at 0		
			group.attrs['total_num_frames'] = total_frames
			if self.sample_dump:
				group['Pilot_Samples'].resize(total_frames, axis=0)
			else:
				group['CSI'].resize(total_frames, axis=0)
			group['RSSI'].resize(total_frames, axis=0)
			group['FrameCompleteTime'].resize(total_frames, axis=0)
			if self.collect_down_rssi:
				group['DownRSSI'].resize(total_frames, axis=0)
			if self.num_uplink_samples:
				group['UplinkData'].resize(total_frames, axis=0)
			#if self.iq_agc_en or self.rssi_agc_en:
			#	self.mob_tx_gain.resize(total_frames, axis=0)

		
	def fix_time_sync(self, duration=2, thresh=.1, user=0):
		""" 
		Take a trace using the current settings and determine if any nodes have bad time_sync based on per-node correlation.
		Shift the ASR for any nodes with a bad sync.
		Note: while it would be nice to have this in argos_init, we need AGC and a trace to run it on, so it is here for now.			
		"""
		from channel_analysis import samps2csi,find_bad_nodes
		
		if duration is not None:
			self.collectCSI(duration=duration)
		csi,iq = self.get_csi()
		userCSI = np.mean(csi[csi.shape[0]//2:,:,:,:,:],2)
		bad_nodes = find_bad_nodes(userCSI, user=user, thresh=thresh)
		[self.bs_nodes[n].basebandCmd('SYNC_DELAY', 1) for n in bad_nodes]
		return bad_nodes
		
	def close(self):
		"""Stop the transport, stop the loop, and close the loop.  Take care of any cleanup."""
		self.stop() # will stop the last capture if it wasn't already. shouldn't be necessary.
		[n.close() for n in self.nodes]
		self.stop_running = True #stop the main thread
		self.thread_loop.join() #probably not necessary.
		self.stop_workers() #should already be stopped, but just in case.
	
	def stop(self):
		if self.h5file is not None and self.h5file.id: #it seems the easy way to check if the h5file is open is just if file.id
			print('Closing h5file %s with %d frames recorded.' % (self.filename, self.max_frame_complete+1) )
			#we could try to save AGC here and cleanup the dataset...
			self.h5file.flush()
			self.h5file.close()

	'''  Setters -- TODO: complete with more setters '''
	
	def set_channel(self, frequency, channel):
		"""Set the channel for trace collection."""
		self.nodes.interfaceCmd('channel', 'RFALL', frequency, channel)
		self.channel = channel
		self.frequency = frequency
	
	def set_tx_gains(self, *tx_gain):
		"""Set the initial TX gains for all nodes."""
		self.nodes.interfaceCmd('tx_gains', 'RFALL', tx_gain)
		self.bs_tx_gain = tx_gain
		self.init_mob_tx_gain = tx_gain
		self.mob_tx_gain = tx_gain	
	
	def set_rx_gains(self, *rx_gain):
		"""Set the initial RX gains for all nodes."""
		self.nodes.interfaceCmd('rx_gains', 'RFALL', rx_gain)
		self.bs_rx_gain = rx_gain	
		self.mob_rx_gain = rx_gain	
		self.target_dB = -self.bs_rx_gain[0] - 14
	
	def set_bs_rx_gains(self, *rx_gain):
		"""Set the RX gains for all base station nodes."""
		self.bs_nodes.interfaceCmd('rx_gains', 'RFALL', rx_gain)
		self.bs_rx_gain = rx_gain	
		self.target_dB = -self.bs_rx_gain[0] - 14

	'''  ***************Plotting code***************  '''
	plot_h5file = None
	csi_filename = None
	def get_h5(self):
		"""Get the most recent log file, open it if necessary."""
		if (not self.plot_h5file) or (self.filename != self.plot_h5file.filename): #if it's closed, e.g. for the C version, open it
			print('Opening %s...' % self.filename)
			self.plot_h5file = h5py.File(self.filename,'r')
		return self.plot_h5file
		
	def get_csi(self,decimate=1):
		"""Get the IQ samples and frequency domain CSI from the last trace.  Mostly used for debug plots."""
		h5f = self.get_h5()
		from channel_analysis import samps2csi
		if not hasattr(self, 'csi') or not hasattr(self, 'iq') or (self.csi_filename != self.filename):
			self.csi,self.iq = samps2csi(h5f['Pilot_Samples'][::decimate], h5f.attrs['num_mob_ant']+1, h5f.attrs['samples_per_user'])
			self.csi_filename = self.filename
		return self.csi,self.iq  #assume if plot_h5file is open there is valid csi,iq
		
		
	def plot_max_iq(self,users=None):
		"""Debug plot useful for checking AGC."""
		import matplotlib.pyplot as plt
		csi,iq = self.get_csi()
		users = users if users is not None else range(iq.shape[1])
		max_iq = np.max(np.max(np.abs(iq), axis=4), axis=2)
		for u in users:		
			plt.figure(60+u,figsize=(10,6))
			plt.clf()
			plt.ylabel('Max IQ of User %d' % u)
			plt.xlabel('Frame')
			for bs in range(max_iq.shape[2]):
				plt.plot(max_iq[:,u,bs])
		
	def plot_ant_mag(self,users=None):
		"""Debug plot useful for checking for bad antennas."""
		import matplotlib.pyplot as plt
		csi,iq = self.get_csi()
		users = users if users is not None else range(iq.shape[1]-1)
		mean_iq = np.mean(np.mean(np.abs(iq), axis=4), axis=2) #both ltss, and across all samples within the frame
		mean_iq_ant = np.mean(mean_iq,axis=0)		#across all frames
		plt.figure(80,figsize=(10,6))
		plt.clf()	
		plt.ylabel('Average Amplitude of Antenna')
		plt.xlabel('BS Antenna')
		for ant in range(mean_iq_ant.shape[0]):
			plt.plot(mean_iq_ant[ant])	
				
	def plot_frame(self,frame=None):
		"""Debug plot of real part of one frame that is useful for checking sync and AGC."""
		import matplotlib.pyplot as plt
		h5f = self.get_h5()
		frame = frame if frame is not None else h5f['Pilot_Samples'].shape[0]//2 #random frame -- let's just choose one in the middle.
		plt.figure(150,figsize=(10,6))
		plt.clf()
		plt.ylabel('Real Part of Frame %d' % frame)
		plt.xlabel('Sample')
		for bs in range(h5f['Pilot_Samples'].shape[1]):
			plt.plot(h5f['Pilot_Samples'][frame,bs,:,0])	
			
	def plot_corr(self,users=None,frame=None,decimate=1):
		"""Debug plot that is useful for checking sync (and general correctness)."""
		from channel_analysis import calCorr
		import matplotlib.pyplot as plt
		csi,iq = self.get_csi(decimate=decimate)
		users = users if users is not None else range(iq.shape[1]-1)
		frame = frame if frame is not None else csi.shape[0]//2 #random frame -- let's just choose one in the middle.
		userCSI = np.mean(csi[:,users,:,:,:],2)
		corr_total,sig_sc = calCorr(userCSI,np.transpose(np.conj(userCSI[frame,:,:,:]),(1,0,2)))		
		plt.figure(160,figsize=(10,6))
		plt.clf()
		plt.ylim([0,1.1])	
		plt.ylabel('Correlation with Frame %d' % frame)
		plt.xlabel('Frame')		
		for u in range(len(users)):	
			plt.plot(corr_total[:,u])

	def plot_gains(self,users=None):
		"""Debug plot of mobile gains that is useful for checking AGC."""
		if not self.iq_agc_en and not self.rssi_agc_en:
			return
		import matplotlib.pyplot as plt
		h5f = self.get_h5()
		users = users if users is not None else range(h5f['mob_tx_gain'].shape[1])

		for u in users:
			plt.figure(170+u,figsize=(10,6))
			plt.clf()
			plt.ylim([-6,32])	
			plt.ylabel('TX Gain of User %d' % u)
			plt.xlabel('Frame')
			zero_idx = np.where(h5f['mob_tx_gain'][:,u,0]==0)[0]
			last_val = h5f['mob_tx_gain'].shape[0] if len(zero_idx) <= 1 else zero_idx[1]  #we have extra 0s on the end we need to truncate
			plt.plot(h5f['mob_tx_gain'][:last_val,u,0],h5f['mob_tx_gain'][:last_val,u,1]) #entry, user, (frame,gain_in_db)
			
	def plot_debug(self):
		"""Plot all debug plots."""
		import matplotlib.pyplot as plt
		plt.close("all")
		self.plot_frame()
		self.plot_corr()
		self.plot_max_iq()
		self.plot_ant_mag()
		self.plot_gains()
		#self.plot_h5file.close()	
		
		
#Example and test code below.  This is not part of argosCC.py:

import sys
import atexit
#import signal

if __name__ == '__main__':

	#cc = ArgosCC(nodesFile='../matlab/config/argos.ips', agc_interval=.05, frame_length = .1e6, iq_agc_en=True, rssi_agc_en=False, collect_down_rssi=False, num_mob_rssi=128, wideband=True, bs_tx_gain=[2,40], bs_rx_gain=[1,5], mob_rx_gain=[1,20], mob_tx_gain=[2,10], both_mob_pilots=False, num_uplink_samples=2048)
	#cc = ArgosCC(nodesFile='../matlab/config/wurc_measurements.ips', agc_interval=.05, frame_length = .1e6, iq_agc_en=True, rssi_agc_en=False, collect_down_rssi=False, num_mob_rssi=128, mob_tx_gain=[25], mob_rx_gain=[12], bs_tx_gain=[25], bs_rx_gain=[20], wurc_channel=563000, both_mob_pilots=False, wideband=True) #wurc
	cc = ArgosCC(nodesFile='./argos.ips', 
		frequency=2.4,
		channel=14, #best 5ghz channels are 5 and 12
		num_uplink_samples=0, #2048 
		agc_interval=0.050, 
		frame_length=.4e6, 
		iq_agc_en=False, 
		rssi_agc_en=True, 
		collect_down_rssi=False,
		num_mob_rssi=128, 
		wideband=True, 
		bs_tx_gain=[29], 
		bs_rx_gain=[38], 
		mob_rx_gain=[52], 
		mob_tx_gain=[25], 
		both_mob_pilots=True,)
		#dont_init=False)
	atexit.register(cc.close) #make sure we save the hdf5 file properly, we don't want to lose a long trace because of a crash. (it also helps to release the udp ports)
	print(cc.nodes)
	#cc.use_CLIB = False
	cc.collectCSI(duration=1, calibrateAoA=False)
	print('cc.collectCSI(duration=10)') #easy copy paste
	
	#drop to command prompt #this really isn't that necessary, since if you launch from an interpreter, it will actually keep the thread running it looks like (likely for the same reason the event loop doesn't catch signals).
	#import code
	#code.interact(local=locals()) #http://stackoverflow.com/questions/2158097/drop-into-python-interpreter-while-executing-function

	#cc.nodes.userExtCmd('sd_reconfig',0)
	#atexit.unregister(cc.close)#so we don't call it twice on accident
	#cc.close()
	#sys.exit(0)
	

	#wireshark filter: ip.src==10.224.76.252 and ip.dst==10.224.76.204 and udp
	#wireshark filter: ip.src==10.224.76.252 and ip.dst==10.224.76.255 and udp
