#******************************************************************************
# 
# File   : argos_init.py
# Authors:	Michael Tsehaie and Clay Shepard (cws [at] rice.edu) (based on Matlab implementation)
# License:	Copyright 2015, Rice Efficient Computing Group. All rights reserved.
#
#******************************************************************************

import warnings
from argoscc.warpNode import *
import argoscc.preambles as Preambles
import numpy as np
import time
import os
import sys
from scipy.linalg import hadamard
from argoscc.argos_writePreamblePageWeights import *
from argoscc.argos_setFrameTiming import *
from argoscc.ArgosBS import ArgosBS
from argoscc.ArgosMobile import ArgosMobile
import h5py

import argoscc.lts as lts

def argos_init(reconfig_nodes='auto', bs_sd_slot=1,mob_sd_slot=1, warp_sd_slot=0,
			   rewrite_sd='none',nodesFile='./argos.ips', bs_bin_path='./bins/ArgosBS.bin',
			   mobile_bin_path='./bins/ArgosMobile.bin', max_corr_length=128, frequency=2.4,
			   channel=14, tx_gains=[20], rx_gains=[40], tx_lpf=3, rx_lpf=3,
			   corr_length=128, cp=0, frame_length=39999, pilot_length=64, sample_rate=20, both_mob_pilots=True,
			   pilot_freq=5e6, dl_buffer_length=512, preamble_type='gold_ifft',
			   corr_threshold=[1.25,1.25], beacon_beamsweep_type='Hadamard',
			   extra_pilots=0, wideband_pilot=False, wurc_tx_gain_db=20, wurc_rx_gain_db=25, wurc_channel=490000, bandwidth=20, 
			   wurc_tx_lpf_mhz=20, wurc_rx_lpf_mhz=20, TDD_guard=128, PilotGuard=80, dont_init=False):  #WURC needs 4 us (80 samples)
	"""
	Initialize all nodes specified in a configuration file with the given parameters.
	
 	The following variables can be set by specifying them as arguments
 	to argos_init.

 	For example, the framelength of BS and Mobile nodes can be set by
	adding the following arguments to the function: ('frame_length', 39999)

	reconfig_nodes	  	Set to:
						0 to skip reconfiguration
						1 to reconfigure all nodes
						'ArgosBS' to configure only BS nodes
						'ArgosMobile' to configure only mobile nodes
	
	rewrite_sd		Set to:
						'none' to skip writing to SDs
						[Node_ID_1, Node_ID_2, ...] to write to the SD cards of nodes specified
	
	nodesFile	 	IPS file
	bs_bin_path		Path to BS bin file (including name of file)
	mobile_bin_path	 	Path to Mobile bin file (including name of file)
	max_corr_length	 	Maximum correlation length as set by hardware file
	frequency		Center TX frequency, e.g. 2.4
	channel			Between 1-14 for 2.4Ghz
	tx_gains		Example: [3,10]
	rx_gains		Example: [1,31]
	tx_lpf			Set to one of [1,2,3] for [24,36,48] MHz of bandwidth
	rx_lpf			Set to one of [0,1,2,3] for [15,19,28,36] MHz of bandwidth
	corr_length		Correlation length size to use
	cp			Cyclic prefix length
	frame_length		Single frame length
	Ts			Baseband clock frequency
	preamble_type	   	Can be set to: 'Gold_IFFT', 'Gold_BPSK', 'Zadoff_Chu'
	corr_threshold	  	Set to size 2 array such as: [2 2]

	--- WURC ADDITION ---

	wurc_tx_gain_db = 20; % [1 30] (try to limit to 20)
	wurc_rx_gain_db = 25; % [0 55] 
	%wurc_channel = 2484000; % 2.4 GHz channel 14 - frequency set in Hz
	wurc_channel = 490000; % 490MHz - frequency set in Hz
	
	% LPF Settings (BY BW): 0-40MHz 1-20MHz 4-10MHz 9-5MHz 
	wurc_tx_lpf_ind = 1;  %tx and rx are the same 9: 5 MHz 4: 10MHz 1: 20 MHz 0: Off
	wurc_rx_lpf_ind = 1;

	"""
	Ts = 1/1e6/sample_rate
	upsample = sample_rate//bandwidth
	wurc_lpf_mhz_ind = {5: 9, 10: 4, 20: 1, 40: 0, 0:0, -1:0} #40 MHz is off, and to be nice, let's let 0 or -1 also be off.
	nodes = initNodes(nodesFile=nodesFile, dont_init=dont_init)
	WarpNode.barrier(nodes)

	bs_nodes, mob_nodes, bswl_nodes, mobwl_nodes, mnode = sortNodes(nodes)

	if isinstance(rewrite_sd,list):
		[n.sdCmd('sd_init') for n in nodes]
		sd_rewrite_nodes = []
		print('Writing to SD nodes (ID): ')
		for n in range(len(rewrite_sd)):
			for node_ind in range(0,len(nodes)):
				if nodes[node_ind].ID == rewrite_sd[n]:		  
					sd_rewrite_nodes.append(nodes[node_ind])
					print('%d ' % nodes[node_ind].ID)		
		[rw_bs, rw_mob] = sortNodes(sd_rewrite_nodes, 'name'); #selective rewrite
		[rwbs.wl_sdCmd('sd_write_bitstream', bs_sd_slot, bs_bin_path) for rwbs in rw_bs]
		[rwmob.wl_sdCmd('sd_write_bitstream', mob_sd_slot, mobile_bin_path) for rwmob in rw_mob]
	elif isinstance(rewrite_sd,str):
		[n.sdCmd('sd_init') for n in nodes]
		if rewrite_sd == 'all':
			rw_bs = bs_nodes + bswl_nodes
			rw_mob = mob_nodes + mobwl_nodes
			print('Writing to SD nodes (ID): ')
			for b_node_ind in range(0,len(bs_nodes)):
				print('%d ' %bs_nodes[b_node_ind].ID)
			for m_node_ind in range(0,len(mob_nodes)):
				print('%d ' %mob_nodes[m_node_ind].ID)
			[rwbs.sdCmd('sd_write_bitstream', bs_sd_slot, bs_bin_path) for rwbs in rw_bs]			  #'C:\MUBF\Code\Argos\BaseStation\build\SDK_Export\ArgosBS_hw\download.bin'		 
			[rwmob.sdCmd('sd_write_bitstream', mob_sd_slot, mobile_bin_path) for rwmob in rw_mob]	  #'C:\MUBF\Code\Argos\Mobile\build\SDK_Export\ArgosMobile_hw\download.bin' 
		elif rewrite_sd == 'none':
			print('No SD cards re-written')
		else:
			raise Exception('String Input Error', 'Could not parse command string.... Only possible command string is ''all'' and ''none''')
	else:
		raise Exception('Input Arg Error', 'Could not parse command argument. rewrite_sd can only be a node number, vector of node numbers, or the string ''all''')


	
	#### Reboot nodes into specified configuration  ####
	#if (isinstance(reconfig_nodes,str) or ((isinstance(reconfig_nodes,int) or isinstance(reconfig_nodes,bool)) and reconfig_nodes) or isinstance(rewrite_sd,int) or rewrite_sd.lower() == 'all'):
	if ( (isinstance(reconfig_nodes,str) and (reconfig_nodes.lower() != 'none') ) or ( (isinstance(reconfig_nodes,int) or isinstance(reconfig_nodes,bool)) and reconfig_nodes) or isinstance(rewrite_sd,int) or rewrite_sd.lower() == 'all'):
		[bs_nodes, mob_nodes, bswl_nodes, mobwl_nodes, mnode] = sortNodes(nodes, 'name')
		[n.sdCmd('sd_init') for n in nodes]
		#print('Warning!!!  SD_Init temporarily disabled!')
		wl_nodes = bswl_nodes +  mobwl_nodes
		wait_for_reconfig = True
		#if reconfigure is set to "auto" we reconfigure only the nodes that need to be
		if (isinstance(reconfig_nodes,str) and reconfig_nodes.lower() == 'auto'):
			wait_for_reconfig = False
			for n in nodes:
				if n.nodetype == n.ARGOSBS_TYPE:
					if (n.name[:7] != 'ArgosBS'):
						n.sdCmd('sd_reconfig', warp_sd_slot) #this is currently an ArgosBS node, but it should be a warplab
						wait_for_reconfig = True
				elif n.nodetype == n.ARGOSMOBILE_TYPE:
					if (n.name[:11] != 'ArgosMobile'):
						n.sdCmd('sd_reconfig', warp_sd_slot) #this is currently an ArgosMobile node, but it should be a warplab
						wait_for_reconfig = True
				elif n.nodetype == n.WARPLAB_TYPE:
					if (len(n.name) > 11 and n.name[:11] =='ArgosMobile'):
						n.sdCmd('sd_reconfig', mob_sd_slot) #this is currently an ArgosMobile node, but it should be a warplab
						wait_for_reconfig = True
					if (n.name[:7] == 'ArgosBS'):
						n.sdCmd('sd_reconfig', bs_sd_slot) #this is currently an ArgosMobile node, but it should be a warplab
						wait_for_reconfig = True
					# If reconfigure node type was specified, reboot only that type
		elif isinstance(reconfig_nodes,str):
			if (reconfig_nodes.lower() == 'argosbs'):
				[n.sdCmd('sd_reconfig', bs_sd_slot) for n in bs_nodes]
			elif (reconfig_nodes.lower() == 'argosmobile'):
				[mob_n.sdCmd('sd_reconfig', mob_sd_slot) for mob_n in mob_nodes]
			elif (reconfig_nodes.lower() == 'wl-mob'):
				[mobwl_n.sdCmd('sd_reconfig', warp_sd_slot) for mobwl_n in mobwl_nodes]
			elif (reconfig_nodes.lower() == 'wl-bs'):
				[bswl_n.sdCmd('sd_reconfig', warp_sd_slot) for bswl_n in bswl_nodes]
			else:  # Error
				raise Exception('Reconfig string input error.reconfig_nodes must either be. an array of indices or a column vector containing one or more of the following: ''ArgosBS'', ''ArgosMobile'', ''WL-BS'', ''WL-MOB')
		else:
			if len(bs_nodes) != 0:
				[n.sdCmd('sd_reconfig', bs_sd_slot) for n in bs_nodes]
			if len(mob_nodes) != 0:
				[n.sdCmd('sd_reconfig', mob_sd_slot) for n in mob_nodes]
			if len(wl_nodes) != 0:
				[n.sdCmd('sd_reconfig', warp_sd_slot) for n in wl_nodes]
				
		if wait_for_reconfig:
			WarpNode.barrier(nodes)
			[n.close() for n in nodes]

			del nodes, bs_nodes, mob_nodes, wl_nodes
			print('Waiting for nodes to reboot...')
			time.sleep(13)		
			print('Done.')
			nodes = initNodes(nodesFile=nodesFile)	 ## check this?
			WarpNode.barrier(nodes) 
		else:
			print('nodes didn\'t need to be reconfigured...')

	[bs_nodes, mob_nodes, bswl_nodes, mobwl_nodes, mnode] = sortNodes(nodes, 'type')
	wl_nodes = bswl_nodes +  mobwl_nodes

	### Configure BS and Mobile nodes ###

	# Generate sine wave data for narrowband pilot
	buffer_length = 512
	time_vals = np.array(np.arange(0,(buffer_length)*Ts,Ts)).transpose()
	sin_data = np.exp(time_vals*1j*2*np.pi*pilot_freq)  # 5 MHz sinusoid as our "payload"

	# Get preambles, ordered according to highest peak to average energy ratio
	#preambles, highest_peak_to_second_ratio = Preambles.getPreambles(preamble_type, corr_length, cp) #don't do this -- it messes up the order
	preambles_bs, highest_peak_to_second_ratio_bs = Preambles.getPreambles(preamble_type, corr_length, cp, upsample=upsample) #the base station may upsample, but the mobiles won't
	preambles = preambles_bs[:,::upsample] #the correlators can run at lower rates, so we only need the downsampled signal.
	
	num_bs_antennas = 4*len(bs_nodes)


	#read precomputed data, or just make it a sine wave.
	data_file_name = 'data/argos_buff_2048_samps_8_antennas.hdf5'
	if os.path.isfile(data_file_name) and len(mob_nodes) <= 8:
	#if False:
		print('Using precomputed data to load uplink and downlink buffers.')
		data_file = h5py.File(data_file_name)
		dl_data = np.asarray(data_file['data_buff']['i']) + np.asarray(data_file['data_buff']['q'])*1j
		dl_data = dl_data/np.abs(dl_data).max()
		#dl_data = np.tile(dl_data,(2,1)) #make sure we have enough
	else:
		dl_time_vals = np.array(np.arange(0,(dl_buffer_length)*Ts,Ts)).transpose()
		dl_data = np.exp(dl_time_vals*1j*2*np.pi*1e6) # 1 MHz sinusoid as our "payload"
		dl_data = np.tile(dl_data,(16,1))
	up_data = dl_data
	

	#note that at one point we did for n in bs_nodes.  Doing each as [n.cmd for n in bs_nodes] or nodes.cmd is better/faster since it rotates commands to each node rather than swamping each node one at a time.
	if len(bs_nodes) > 0:		
			
		[n.basebandCmd('ENSYNCOUT',0) for n in bs_nodes] # Disable continuous frames for BS nodes
		
		# Load beacon and paging sequences
		[[n.basebandCmd(n.baseband.TXBUFF_PAGE[preamble_index], preambles_bs[preamble_index, :]) for n in bs_nodes] for preamble_index in range(min(len(mob_nodes),len(ArgosBS.TXBUFF_PAGE)))] #todo: support both ArgosBS and ArgosBS_WURC (for maximum number of users)
		
		# Write DL data
	
		[[n.basebandCmd(n.baseband.TXBUFF_TERM[i], dl_data[i]) for n in bs_nodes] for i in range(4)] #range(min(len(mob_nodes),len(ArgosBS.TXBUFF_PAGE)))] #todo: support more ArgosBS and ArgosBS_WURC
		#[n.basebandCmd(ArgosBS.TXBUFF_TERM[i], [dl_data]) for i in range(len(ArgosBS.TXBUFF_TERM))]
					#for k in range(len(bs_nodes)):
						#bs_nodes[k].basebandCmd(ArgosBS.TXBUFF_TERM(i), [[dl_data]])
		
		# Set BS preamble sweep params
		[n.basebandCmd('NUMBS', num_bs_antennas) for n in bs_nodes]
		[n.basebandCmd('NUMBEAMPERSWEEP', num_bs_antennas) for n in bs_nodes]
		[n.basebandCmd('DISABLE_CALIB') for n in bs_nodes]
		[n.basebandCmd('DISABLE_SCALING') for n in bs_nodes]
		[n.basebandCmd('DISABLE_TERM_SCALING') for n in bs_nodes]
		[n.basebandCmd('ENABLE_BYPASSMOD') for n in bs_nodes]
		[n.basebandCmd('DISABLE_USEFFT') for n in bs_nodes]
		[n.basebandCmd('PAGE_SEL', 0) for n in bs_nodes]
		[n.basebandCmd('PAGE_BW_RST', 1) for n in bs_nodes]
		[n.basebandCmd('TOGGLENFRAMES', 2**32-1) for n in bs_nodes]
		
		# Set beamforming weights

		if beacon_beamsweep_type == 'Hadamard':
			possible_dim = []
			possible_dim.append(2**(np.ceil(np.log2(num_bs_antennas))))
			#scipy doesn't support these! (though it should...)
			#possible_dim.append(20*2**(np.ceil(np.log2(np.ceil(num_bs_antennas/20)))))
			#possible_dim.append(12*2**(np.ceil(np.log2(np.ceil(num_bs_antennas/12)))))
			h_dim = min(possible_dim)
			hadamard_matrix = hadamard(h_dim)	#hadamard matrix : http://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.linalg.hadamard.html
			beacon_weights = hadamard_matrix[0:num_bs_antennas, 0:num_bs_antennas]
		else:
			beacon_weights = np.identity(4*len(bs_nodes))
		
		# Disable paging by default
		page_weights = np.zeros((4*len(bs_nodes),4*len(bs_nodes)))
		
		# Currently send beacon/paging sequence from single antenna at a time
		#beacon_weights = np.identity(4*len(bs_nodes))
			#page_weights = np.zeros((4*len(bs_nodes),4*len(bs_nodes)))
		#print(beacon_weights)
		argos_writePreamblePageWeights(bs_nodes, (1-2**-15)*beacon_weights, (1-2**-15)*page_weights)
		[n.basebandCmd('PREAMBLE_BW_RST', 0) for n in bs_nodes]
		
		# Set calibration pilot indices for BS nodes
		
		# Load pilots
		[n.basebandCmd('TXBUFF_PILOT', sin_data) for n in bs_nodes]
		[n.basebandCmd('INVPILOT_BUFF', (1/sin_data)) for n in bs_nodes]
		#print ("Finished BS Initialization!")
	for i in range(len(bs_nodes)):
		#bs_nodes[i].basebandCmd('BS_INDEX_RADIO', np.arange((i+1-1)*4+1,(i+1-1)*4+5))
		for j in range(4):
			bs_nodes[i].basebandCmd(bs_nodes[i].baseband.BS_INDEX_RADIO[j], i*4+j+1) 
			#[n.basebandCmd(n.baseband.BS_INDEX_RADIO[j], i*4+j+1) for n in bs_nodes] #have to use n.baseband, since we can have different basebands...
	
	beacon = preambles[0,cp+corr_length:] 
	
	if len(mob_nodes) > 0:
		# Set mobile corr thresholds
		[n.basebandCmd('CORR_0_SET_THRESHOLD', int(corr_threshold[0]*2**8)) for n in mob_nodes]
		[n.basebandCmd('CORR_1_SET_THRESHOLD', int(corr_threshold[1]*2**8)) for n in mob_nodes]
		
		# Disable continuous frames
		[n.basebandCmd('CONTINUOUSFRAMES', 0) for n in mob_nodes]
		[n.basebandCmd('CORR_0_BUFF', beacon.conjugate() ) for n in mob_nodes]
		
		# Set mobiles to use radio A
		[n.basebandCmd('SET_CORR_RADIO_TO_A') for n in mob_nodes]
		
		# Disable CFO and DCO
		
		[n.basebandCmd('DISABLE_CFO') for n in mob_nodes]
		[n.basebandCmd('SET_CFO_HIGH_TRESH', 128) for n in mob_nodes]
		[n.basebandCmd('SET_CFO_LOW_TRESH', 1) for n in mob_nodes]
	
		[n.basebandCmd('DISABLE_TIMING_CORRECTION', 1) for n in mob_nodes]
	
		[n.basebandCmd('DISABLE_CONTINUOUS_DCO') for n in mob_nodes]
		[n.basebandCmd('DISABLE_DCO') for n in mob_nodes]
		#n.basebandCmd('SETDCO', [mob_nodes[0].basebandCmd('FRAMELENGTH')+666, blocking = True])  #todo: michael ... why not just use the framelength supplied?  If this is how matlab does it, it needs to be fixed there too :/
					#n.basebandCmd('SETDCO', [frame_length])
		
		[n.basebandCmd('SET_DCO_LENGTH', 2) for n in mob_nodes]
		
		# Set AGC variables
		[n.basebandCmd('SET_AGC_LENGTH', 2) for n in mob_nodes]
		#n.basebandCmd('SETAGC', [mob_nodes[0].basebandCmd('FRAMELENGTH')+100])
	
		# Set correlator length  
		#[n.basebandCmd('SET_CORR_LENGTH', int(np.log2(len(preambles[0,cp+corr_length:preambles.shape[1]]))-5)) for n in mob_nodes]
		[n.basebandCmd('SET_CORR_LENGTH', int(np.log2(corr_length)-5)) for n in mob_nodes] #-5 is the offset specific to the correlator core (0 is 32 length, 1 is 64, 2 is 128)
		[n.basebandCmd('SET_CORR_RATE', int(np.log2(upsample))) for n in mob_nodes]
		
		# Set and load uplink data.
		#up_data = sin_data 
		if both_mob_pilots:
			[n.basebandCmd('TXBUFF_RADIO_1', up_data[i*2]) for i,n in enumerate(mob_nodes)]
			[n.basebandCmd('TXBUFF_RADIO_2', up_data[i*2+1]) for i,n in enumerate(mob_nodes)]
		else:
			[n.basebandCmd('TXBUFF_RADIO_1', up_data[i]) for i,n in enumerate(mob_nodes)]
			[n.basebandCmd('TXBUFF_RADIO_2', np.zeros(dl_buffer_length)) for n in mob_nodes] #todo: make ul_buffer_length parameter
	
		# Load pilots onto mobiles
		if wideband_pilot:			 
			wb_pilot = np.concatenate((np.tile(lts.genLTS(upsample=upsample,cp=0),4),np.zeros(512, dtype='complex64')))
			[n.basebandCmd('TXBUFF_PILOT', wb_pilot[:512] ) for n in mob_nodes]	
		else:
			[n.basebandCmd('TXBUFF_PILOT', sin_data) for n in mob_nodes]
	
		# Rst correlators at end to take care of changes in rate/length etc.
		# w.r.t default values

		[n.basebandCmd('RST_CORR') for n in mob_nodes]
	
		[n.basebandCmd('RESETFRAMECOUNT', 1) for n in mob_nodes]
		[n.basebandCmd('RESETFRAMECOUNT', 0) for n in mob_nodes]
		#WarpNode.barrier(nodes) 
		# Set pilot indices for mobile nodes
	for i in range(len(mob_nodes)): #todo, use mobile index and move this to the main for loop
		mob_nodes[i].basebandCmd('CORR_1_BUFF', preambles[1+i, cp+corr_length:].conjugate() )
		if (both_mob_pilots):  #todo: this should also set the frametiming... but oh well.
			mob_nodes[i].basebandCmd('RADIO_1_TERMINAL_INDEX', 2*i)
			mob_nodes[i].basebandCmd('RADIO_2_TERMINAL_INDEX', 2*i+1)
		else:
			mob_nodes[i].basebandCmd('RADIO_1_TERMINAL_INDEX', i)
			mob_nodes[i].basebandCmd('RADIO_2_TERMINAL_INDEX', 2*len(mob_nodes)+i+1+2) #sometime after the pilot phase, so this should never be reached
		#print ("Finished Mobile Initialization!")
	   
	# Set frame timing
	argos_setFrameTiming(bs_nodes + mob_nodes,  PreambleLength=np.floor(1.5*preambles.shape[1]), FrameLength=frame_length, PilotLength=pilot_length, extra_pilots=extra_pilots, TDD_guard=TDD_guard, both_mob_pilots=both_mob_pilots, PilotGuard=PilotGuard)
	#argos_setFrameTiming(bs_nodes + mob_nodes, DownLength=128, UpLength=128, PreambleLength=np.floor(1.5*preambles.shape[1]), FrameLength=frame_length, PilotLength=pilot_length, extra_pilots=extra_pilots, TDD_guard=TDD_guard, both_mob_pilots=both_mob_pilots, PilotGuard=PilotGuard)
				
				
	##### Common configuration parameters  #######

	#Normal Maxim 2829 Interface Configuration
	#set AGC parameters
	[n.interfaceCmd('rx_gain_mode', n.RFALL_M2829, 'manual') for n in nodes] #bs_nodes may not have this command... (this build does, but if not, it should be commented out, and this should be set in the individual sections).

	# Set TX/RX parameters
	[n.interfaceCmd('channel', n.RFALL_M2829, frequency, channel) for n in nodes]
	[n.interfaceCmd('tx_gains', n.RFALL_M2829, *tx_gains) for n in nodes]
	[n.interfaceCmd('rx_gains', n.RFALL_M2829, *rx_gains) for n in nodes]

	#Set filter bandwidth
	[n.interfaceCmd('tx_lpf_corn_freq', n.RFALL_M2829, tx_lpf) for n in nodes] #Configure Tx in [1,2,3] for [24,36,48]MHz of bandwidth
	[n.interfaceCmd('rx_lpf_corn_freq', n.RFALL_M2829, rx_lpf) for n in nodes] #Configure Rx in [0,1,2,3] for [15,19,28,36]MHz of bandwidth
	
	#WURC Interface Configuration
	#WURC SPECIFIC PARAMETERS (ONLY FOR BS NODES)
	bs_nodes.interfaceCmd('channel', 'RFALL_LMS6', wurc_channel)
	bs_nodes.interfaceCmd('rx_gains', 'RFALL_LMS6', wurc_rx_gain_db)
	bs_nodes.interfaceCmd('tx_gains', 'RFALL_LMS6', wurc_tx_gain_db)

	bs_nodes.interfaceCmd('tx_lpf_corn_freq', 'RFALL_LMS6', wurc_lpf_mhz_ind[wurc_tx_lpf_mhz]) # See param setting above for codes
	bs_nodes.interfaceCmd('rx_lpf_corn_freq', 'RFALL_LMS6', wurc_lpf_mhz_ind[wurc_rx_lpf_mhz])
	bs_nodes.interfaceCmd('rx_en', 'RFALL_LMS6')  #naren had this, but it doesn't seem like we should need it...

	
	#WURC SPECIFIC PARAMETERS (ONLY FOR BS NODES)
	mob_nodes.interfaceCmd('rx_gains', 'RFALL_LMS6', wurc_rx_gain_db)
	mob_nodes.interfaceCmd('tx_gains','RFALL_LMS6', wurc_tx_gain_db)
	mob_nodes.interfaceCmd('channel', 'RFALL_LMS6', wurc_channel)


	mob_nodes.interfaceCmd('tx_lpf_corn_freq', 'RFALL_LMS6', wurc_lpf_mhz_ind[wurc_tx_lpf_mhz]) # See param setting above for codes
	mob_nodes.interfaceCmd('rx_lpf_corn_freq', 'RFALL_LMS6', wurc_lpf_mhz_ind[wurc_rx_lpf_mhz])

	mob_nodes.interfaceCmd('rx_en', 'RFALL_LMS6') #naren had this, but it doesn't seem like we should need it...
	return nodes, mnode, bs_nodes, mob_nodes, bswl_nodes, mobwl_nodes


import signal
import sys
import threading

if __name__ == '__main__':
	#nodes = initNodes(nodesFile='./argos.ips')
	#[nodes, mnode, bs_nodes, mob_nodes, bswl_nodes, mobwl_nodes] = argos_init(nodesFile='../matlab/config/wurc_measurements.ips')
	[nodes, mnode, bs_nodes, mob_nodes, bswl_nodes, mobwl_nodes] = argos_init(nodesFile='../../matlab/config/argos_automtu.ips', rewrite_sd = 'none')
	#[nodes, mnode, bs_nodes, mob_nodes, bswl_nodes, mobwl_nodes] = argos_init(nodesFile='../matlab/config/argos.ips', wideband_pilot=True, tx_gains=[1,10], rx_gains=[1,1])
	nodes.barrier()
	print(nodes)
	#nodes.sdCmd('sd_reconfig',0)
	import code
	code.interact(local=locals())

	sys.exit()
