#******************************************************************************
# 
# File   : interface_m2829.py
# Authors:	Clay Shepard (cws [at] rice.edu)
# License:	Copyright 2015, Rice Efficient Computing Group. All rights reserved.
#
#******************************************************************************

class InterfaceM2829():
	"""Class for default 2.4/5 GHz Maxim 2829 radio interfaces."""

	description = 'Maxim 2829 Interface.'

	tx_en = 1
	rx_en = 2
	tx_rx_dis = 3
	channel = 4
	tx_gains = 5
	rx_gains = 6
	tx_lpf_corn_freq = 7
	rx_lpf_corn_freq = 8
	rx_gainctrl_src = 9
	rx_gain_mode = 9
	
	manual = 0
	automatic = 1
	
	#disable = 0
	#enable = 1

	tx_gain_max 		= 31;
	tx_gain_min 		= -5;
	rx_gain_max 		= 93;
	rx_gain_min 		= 0;

	#This will interpret dB to maxim gains -- todo, add this to the C interface
	#Todo allow relative gain increase to C code.
	@classmethod	
	def procCmd(cls, cmdID, *args):
		"""Convert a command in to the proper command ID and data."""
		if type(cmdID) is str:
			cmdID = getattr(cls,cmdID.lower())
		
		args = list(args)
		if cmdID == cls.tx_gains:
			if len(args) == 1:
				args = cls.db2txgain(args[0])
			elif len(args) == 2:
				print('Warning! 2 parameter gains are deprecated for Maxim 2829.  Please use dB setting.')
			if len(args) >= 2 and ( args[0] < 0 or args[0] > 3 or args[1] < 0 or args[1] > 63):  #bad gain setting
				print('Invalid TX Gains! %s.' % (str(args)) )
				return
		
		if cmdID == cls.rx_gains:
			if len(args) == 1:
				args = cls.db2rxgain(args[0])
			elif len(args) == 2:
				print('Warning! 2 parameter gains are deprecated for Maxim 2829.  Please use dB setting.')
			if len(args) >= 2 and ( args[0] < 1 or args[0] > 3 or args[1] < 0 or args[1] > 31): #bad gain setting
				print('Invalid RX Gains! %s.' % (str(args)) )
				return
		
		if cmdID == cls.channel:
			if args[0] == 2.4:
				args[0] = 0
			elif args[0] == 5:
				args[0] = 1
			else:
				raise Exception("Bad frequency chosen: expects band selection 2.4 or 5.")
				return

		#todo: probably remove this.  it just checks 'manual' vs. 'automatic', which we should check explicitly.
		for i in range(len(args)):
			if type(args[i]) is str:
				args[i] = getattr(cls,args[i].lower())	# manual vs automatic and disable enable.

		return cmdID, args

		
	tx_bb_step = [-5, -3, -1.5, 0]
	tx_rf_step = 31.0/63 #0.5	
	@classmethod
	def txgain2db(cls, bb, rf):		
		return cls.tx_bb_step[bb] + cls.tx_rf_step*rf
	
	@classmethod
	def db2txgain(cls, dB):
		"""
		Convert the desired TX gain (dB) to the appropriate Maxim 2829 gains.
		Valid range is from -5 dB to 31 dB.
		The power scale is somewhat arbitrary, but at max gain the output should be at least 20 dBm.
		"""
		#max dB we can support is 63*tx_rf_step = 31, min is tx_bb_step[0] = -5
		bb = 0
		rf = 0
		if dB > 23:
			bb = 3
		elif dB > 15:
			bb = 2
		elif dB > 7:
			bb = 1
		
		rf = (dB-cls.tx_bb_step[bb])/cls.tx_rf_step
		if rf < 0:
			print('Warning! %d dB below TX gain range. [%i,%i]' % (dB,cls.tx_gain_min,cls.tx_gain_max) )
			rf = 0
		if rf > 63:
			print('Warning! %d dB above TX gain range. [%i,%i]' % (dB,cls.tx_gain_min,cls.tx_gain_max) )
			rf = 63
			
		return [int(bb),int(rf)]
		
	rx_rf_step = 15
	rx_bb_step = 2 #63/31
	
	@classmethod
	def rxgain2db(cls, rf, bb):	 #these are intentionally switched, as maxim is assymmetric (3 bb rf rx steps, but 3 bb rf tx steps), and to be consistent with WARP commands
		return cls.rx_bb_step[bb] + cls.rx_rf_step*rf

	@classmethod
	def db2rxgain(cls, dB):
		"""
		Convert the desired RX gain (dB) to the appropriate Maxim 2829 gain settings.
		Valid range is from 0 dB to 93 dB.
		"""
		#scale goes from 0 to 93 (31*2+2*15)
		#todo:scale dB based on actual dBm
		
		rf = 1
		if dB > 60:
			rf = 3
		elif dB > 30:
			rf = 2
		
		bb = (dB-cls.rx_rf_step*(rf-1))/cls.rx_bb_step
		if bb < 0:
			print('Warning! %d dB below RX gain range. [%i,%i]' % (dB,cls.rx_gain_min,cls.rx_gain_max) )
			bb = 0
		if bb > 31:
			print('Warning! %d dB above RX gain range. [%i,%i]' % (dB,cls.rx_gain_min,cls.rx_gain_max) )
			bb = 31
			
		return [int(rf),int(bb)]
		
	@classmethod
	def target_dB_fromRxGain(cls, rf, bb, target=0):
		"""Target receive dB based on rx gains settings"""
		return -cls.rxgain2db(rf,bb) + 0 + target #This is based on an LTS and some testing.  It seems reasonable.
	
	@classmethod	
	def targetRSSI(cls, bb, rf):
		'''RSSI we want to hit given a rx gain setting (on the BS)'''
		return 1024 - (cls.rx_rf_step*rf + cls.rx_bb_step*bb)*1024/70  #since the rf gain setting changes the RSSI, do we even need to change it here?
	
	@staticmethod
	def rssi2dB(rssi, rx_rf):
		"""Takes rssi (from 0 to 1023) and the receiver's rf gain setting and returns the received signal strength in dBm."""
		return rssi*70/1023 - 55 - rx_rf*15