#!/usr/bin/python3
#******************************************************************************
# 
# File   : phase_calibrator.py
# Authors:	Andrew Brooks (agb3 [at] rice.edu)
# License:	Copyright 2015, Rice Efficient Computing Group. All rights reserved.
#
#******************************************************************************

"""
This is a calibration program that can either be run as a command line utility
or included as a library in any collection utilities.

Here is the basic idea behind calibration:

- Send/Receive one LTS from directly in front of the antenna array. If
	possible, use a directional antenna. Set aside this LTS.
- Then, run your experiment as you would ordinarily, pretending that the PLLs
	don't have a random phase offset.
- Because the position is known in the first one, the phase on every subcarrier
	at every antenna should be zero. So, we use it to correct all of the
	subsequent measurements by effectively subtracting out that phase.

- Save your data in such a way so as to be compatible with the calibrator
	(see below), or use functions from this file in your code.
	(also see below).
	

ASSUMPTIONS FOR USE AS A STANDALONE UTILITY:

1. Your data is saved inside an HDF5 file with the following properties:

	a. The signal's real part is saved as the '/r' dataset.
		- This dataset is a numpy matrix of dimensions:

		[array x-size, array y-size, signal length].
	b. The signal's imag part is saved as the '/i' dataset.
		Dataset properties are the same as /r.
	c. The LTSes are "close-packed" -- every 160 samples is another LTS.

2. Your calibration is saved in an HDF5 file where:
	a. (a.) and (b.) from above are true
	b. You have only one LTS in the entire file.
	c. That LTS was sent from one antenna several feet in front the array,
		very preferably from a directional antenna facing the array.

3. If your data doesn't meet these assumptions or you just want to do this
	automatically, just `import calibrate` from a file in the same dir, and
	then you can use these functions.

INVOCATION

	./phase_calibrator.py calibration_lts.hdf5 data.hdf5 calibrated.hdf5

	This will do the following:

	i. Get the phases for each subcarrier in calib.hdf5.

		(ie, FFT(LTS) -> matrix of phase vectors)
	ii. For every LTS in data.hdf5:
		a. Take the FFT.
		b. Multiply every i-th subcarrier by exp(-j*phase[i])
			(at every antenna).
		c. IFFT, re-add cyclic prefix, then store in calibrated.hdf5.

Questions? --> abrooks@rice.edu
"""

import argoscc.lts
import h5py
import numpy as np 

SC_INDS = np.concatenate((np.arange(2, 54, 2), -1 * np.arange(2, 54, 2)))

def load_iq(hfile):
	"""
	Loads HDF5 data from a named file, returns a numpy matrix with the IQ data.
	"""
	try:
		f = h5py.File(hfile, 'r')
	except:
		raise Warning("Could not open %s." % hfile)
	try:
		i = np.array(f['i'])
		r = np.array(f['r'])
		data = r + 1j*i
	except:
		raise Warning("Input %s is improperly formatted, and does not \
have a /r and /i dataset." % hfile)
	return data

def get_calib_mtx(calib_data):
	"""
	Gets calibration phases from given calibration data. Returns calib_phases,	
	a m-by-n-by-52 matrix for an array of dimensions m-n.
	""" 
	size = np.shape(calib_data)
	ltses = np.fft.fft(calib_data[:,:,32:160])
	calib_phases = np.angle(ltses[:,:,SC_INDS])
	return calib_phases

def correct_ltses(mtx, calib_phases):
	"""
	Performs a correction of mtx by complex-dividing the phases for each
	antenna and subcarrier with calibtration phases. "unzips" mtx with an fft,
	complex divides, iffts back, adding a cyclic prefix for completeness.
	"""
	size = np.shape(mtx)
	calibrated = np.empty(size, dtype='complex64')
	for x in range(0, size[2], 160):
		ltses = np.fft.fft(mtx[:,:,(x+32):(x+160)])
		#Complex divide out calibration phases
		ltses[:,:,SC_INDS] *= np.exp(-1j*calib_phases)
		ltses = np.fft.ifft(ltses, axis=2)
		#Re-add cyclic prefix:
		ltses = np.dstack((ltses[:,:,-32:], ltses))
		calibrated[:,:,x:x+160] = ltses
	return calibrated

def write_iq(calibrated_mtx, hfile):
	"""
	Returns a data matrix to a file.
	"""
	f = h5py.File(hfile, 'w')
	reals = f.create_dataset('r', np.shape(calibrated_mtx))
	imags = f.create_dataset('i', np.shape(calibrated_mtx))
	reals[...] = np.real(calibrated_mtx)
	imags[...] = np.imag(calibrated_mtx)

if __name__ == "__main__":
	import sys
	print('\n')
	if len(sys.argv) < 4:
		raise Warning ("Not enough arguments.")
	elif len(sys.argv) > 4:
		print("[WW] -- You specified too many arguments, but we'll proceed.")
	print("Calibration dataset:  	%s" % sys.argv[1])
	print("Uncalibrated dataset: 	%s" % sys.argv[2])
	print("Output dataset:       	%s" % sys.argv[3])
	(c_name, uc_name, o_name) = sys.argv[1:4]
	c_data = load_iq(c_name)
	uc_data = load_iq(uc_name)
	print("\nProcessing datasets...")
	phase_mtx = get_calib_mtx(c_data)
	correct = correct_ltses(uc_data, phase_mtx)
	write_iq(correct, o_name)
	print("Done.")
