#******************************************************************************
# 
# File   : warpNode.py
# Authors:	Clay Shepard (cws [at] rice.edu)
# License:	Copyright 2015, Rice Efficient Computing Group. All rights reserved.
#
#******************************************************************************

import sys
import asyncio
import struct
import numpy as np
import warnings
import time
import configparser 
import logging
import os
from argoscc.ArgosBS import ArgosBS
from argoscc.ArgosBS_WURC import ArgosBS_WURC
from argoscc.ArgosMobile import ArgosMobile
from argoscc.wl_baseband import wl_baseband
from argoscc.interface_m2829 import InterfaceM2829
from argoscc.interface_lms6 import InterfaceLMS6
from argoscc.sd_cfg import sd_cfg
from argoscc.common_util import CallableList
import netifaces as nif
import concurrent
import threading
import socket

#NOTE: for some reason in windows sometimes it won't receive replies.  I had to reboot to fix it.
#NOTE: I noticed that if you send *a lot* of commands back to back to the same node there are some weird errors that look like "python protocol.pause_writing() failed".  Perhaps we need to throttle somehow?
#TODO: rename file to warp_node.py to follow python convention.

def initNodes(nodesDicts=None,numNodes=None,nodesIDs=None,nodesFile='./argos.ips',dont_init=False):
	"""
	Initialize an array of WarpNode objects.
	
	They can be initialized automatically by specifying a number of nodes (configured sequentially by dipswitches),
	by specifying the node IDs (and thus IPs, configured by dipswitches),
	by a nodesDict, which is an array of dicts and must have at least an ID and an IP defined for each node,
	or by a config file in the standard WARP/Argos format.	
	Make sure you have a large MTU for best performance. (e.g. sudo ifconfig eth0 mtu 9000)
	Note: In Windows you must run:

		netsh interface ipv4 set subinterface "Local Area Connection" mtu=9100 store=persistent
		netsh interface ipv4 show subinterfaces
		
		To make it persist: click start type regedit the navigate to [HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services\Tcpip\Parameters\Interfaces\[Adapter ID]]
	"""

	nodes = CallableList() 
	if numNodes is not None:
		for n in range(numNodes):
			nodes.append(WarpNode(ID=n, IP=('10.224.76.%d' % n), name=('WL-BS_AutoNode_%d' % n)) )
	if nodesIDs is not None:
		for n in nodesIDs:
			nodes.append(WarpNode(ID=n, IP=('10.224.76.%d' % n), name=('WL-BS_AutoNode_%d' % n)) )
	if nodesDicts is not None:
		for n in nodesDicts:
			nodes.append(WarpNode(*n))
	if nodesFile is not None:
		with open(nodesFile) as f:
			next(f) #skip header			
			for line in f:
				if line.find('#') == -1: # not found
					fields = line.split(',')
					if len(fields) >= 4:
						serial = int(fields[0].strip().lower().replace('w3-a-',''))
						id = int(fields[1].strip())
						ip = fields[2].strip()
						name = fields[3].strip()
						nodes.append( WarpNode(ID=id, IP=ip, name=name, serial=serial, dont_init=dont_init) ) #dont_init is temporary hack for 5GHz not syncing right
	nodes.barrier() #make sure they have all been initialized
	return nodes

	
def sortNodes(nodes, method='type', indexNodes=True):
	"""
	Sort an array of WARP nodes based on either their type or their name.
	Returns (bs_nodes, mob_nodes, bswl_nodes, mobwl_nodes, mnode)
	"""
	bs_nodes = CallableList()
	mob_nodes = CallableList()
	bswl_nodes = CallableList()
	mobwl_nodes = CallableList()
	mnode = None
	
	if method.lower() ==  'type':	
		for n in nodes:
			#n.logger.info('node %s, %s, has type %s (%s)' % (str(n.ID), str(n.name), str(n.nodetype), n.description) ) #n._TYPE_NAMES_LONG[n.nodetype]))
			if n.name.lower().find('master') != -1:
				mnode = n
			elif n.nodetype & n.ARGOSBS_TYPE: #bitmask now, since WURC
				bs_nodes.append(n)
				if n.version_hash != bs_nodes[0].version_hash:
					print ('According to the version hash, node %d has a different version from others!' %n.ID)
			elif n.nodetype & n.ARGOSMOBILE_TYPE:
				mob_nodes.append(n)
				if n.version_hash != mob_nodes[0].version_hash:
					print ('According to the version hash, node %d has a different version from others!' %n.ID)
			elif n.nodetype == n.WARPLAB_TYPE:
				if n.name.lower().find('wl-bs') != -1:
					bswl_nodes.append(n)
				elif n.name.lower().find('wl-mob') != -1:
					mobwl_nodes.append(n)
				else:
					n.logger.warn('Warning! node %d, named %s, is configured as WARPLab, but does not have a valid name.  Defaulting to BS.' % (n.ID, n.name) )
					bswl_nodes.append(n)					
		if mnode is not None:
			if mnode.nodetype & n.ARGOSBS_TYPE:
				bs_nodes.insert(0, mnode)
				if len(bs_nodes) > 0 and mnode.version_hash != bs_nodes[0].version_hash:
					print ('According to the version hash, node %d has a different version from others!' %n.ID)
			elif mnode.nodetype == n.WARPLAB_TYPE:
				bswl_nodes.insert(0, mnode)
				
	elif method.lower() == 'name':			
		for n in nodes:
			if n.name.lower().find('master') != -1:
				mnode = n
			elif n.name.lower().find('argosbs') != -1:
				bs_nodes.append(n)
			elif n.name.lower().find('argosmobile') != -1:
				mob_nodes.append(n)
			elif n.name.lower().find('wl-bs') != -1:
				bswl_nodes.append(n)
			elif n.name.lower().find('wl-mob') != -1:
				mobwl_nodes.append(n)	
		if mnode is not None:
			if mnode.name.lower().find('argosbs') != -1:
				bs_nodes.insert(0, mnode)
			elif mnode.name.lower().find('wl-bs') != -1:
				bswl_nodes.insert(0, mnode)
	
	#index the nodes so we can use their index for arrays.
	if indexNodes:
		idx = 0
		for n in bs_nodes:
			n.idx = idx
			idx += 1
		idx = 0
		for n in mob_nodes:
			n.idx = idx
			idx += 1	
	
	return (bs_nodes, mob_nodes, bswl_nodes, mobwl_nodes, mnode)

class WarpNode: #object #DatagramProtocol
	"""
	The primary WarpNode class which represents a single WarpNode, and controls all communication with that node.
	Use "WarpNode(discover_nodes=True)" to print out a list of nodes on the network.	
	"""
	
	#all of these probably can be omitted since they are set in __init__ and info_callback, but I like having all of the class variables outlined.
	ID = None								#unique identification for this node
	description = 'Uninitialized WARP Node' #string description of this node (auto-generated)
	serial = None							#node's serial number, read from EEPROM on hardware
	eth_MAC_addr = None						#node's MAC address (for ETH_A on WARP v3)
	hwVer = None							#WARP hardware version of this node
	wlVer_major = None						#WARPLab version running on this node
	wlVer_minor = None						#
	wlVer_revision = None						#
	num_interfaces = 0						#vector of length num_interfaceGroups with
	baseband = None							#node's baseband object
	nodetype = None							#Basestation node (1) or terminal node (2)
	version_hash = None		 				#Hash that should be unique to the build of the bitstream running on the WARP board -- used to easily identify if different boards are running different versions.
	name = None

	inited = False
	
	#Command Groups - Most Significant Byte of Command ID
	_GRPID_NODE = 0x00
	_GRPID_TRANS = 0x10
	_GRPID_IFC = 0x20
	_GRPID_BB = 0x30
	_GRPID_TRIGMNGR = 0x40
	_GRPID_USER = 0x50
	_GRPID_ARGOSBS = 0x60
	_GRPID_ARGOSMOBILE = 0x70
	_GRPID_WSD = 0x80  #WURC WSD LMS6
	
	#Their C counterparts are found in wl_node.h
	#user friendly
	initialize = 1
	info = 2
	identify = 3
	temperature = 4 #max([((t[0]/65536.0)/0.00198421639) - 273.15 for t in x])
	setup = 5
	reset = 6
	set_cc_macip = 7
	dump_pilot_buff = 8
	dump_up_buff = 9
	
	#These constants define the nodetypes
	#They require the corresponding nodes to report the correct type in 'get_hardware_info' response
	#If the node doesn't report a type, it will default to WARPLAB_TYPE
	WARPLAB_TYPE = 0
	ARGOSBS_TYPE = 1
	ARGOSMOBILE_TYPE = 2
	WURC_TYPE = 4
	_TYPE_NAMES = ['WL', 'BS', 'MOB','bad','bad','BSW','MOBW'] #used for printing out the node.  (It's a bitmask, so this is a lookup table, the bads are the WURC warplabs, which aren't built yet)
	_TYPE_NAMES_LONG = ['WARPLab', 'ArgosBS', 'ArgosMobile','bad','bad','ArgosBS_WURC','ArgosMobile_WURC'] #used for printing out the node.
	
	#todo:subclass interface?
	RFA = 0x10000000
	RFB = 0x20000000
	RFC = 0x40000000
	RFD = 0x80000000
	RFS = [RFA, RFB, RFC, RFD]
	RFALL = 0x00000000 #this is initialized in the hw info callback -- by default there are no interfaces.

	_unicast_starting_port = 9000	
	_bcast_port = 10000
	

	class WarpCommand:
		""""A command that can send itself, as well as track how many times it was sent."""
		def __init__(self, node, seq_num, cmdID, args=None, args_message=None, nodeID=None, bcast=False, flags=1, callback=[]):

			#required arguments.
			self.node = node
			self.seq_num = seq_num
			self.cmdID = cmdID

			#optional arguments
			self.args = args
			self.resend_counter = 0 # a counter that records the time this command has been resent
			self.bcast = bcast
			self.flags = flags
			self.callback = callback
			
			self.finished = False #flag for deletion
			
			#build message based on args.  If args are provide, turn them in to a byte string. If a byte string is already provided, figure out it's length:			
			if args is not None:
				numArgs = len(self.args)
				#print(args)
				args_message = struct.pack(">%dI" % numArgs, *args)
			elif args_message is not None:
				numArgs = len(args_message)/4 #do we need to make sure it's a multiple of 4, and pad if not?
			else:
				args_message = b''
				numArgs = 0
			
			cmdLength = numArgs*4  #Is numArgs ever not cmdLength*4?  If not, maybe we should be able to pass the numArgs in the args_message...
			length = cmdLength + 8
			type = 1 #todo, what is this?
			ID =  self.node.ID if not self.bcast else 0xFFFF  #for broadcast this has to be FFFF
			byte_array = [ID, self.node._host_ID, type, length, self.seq_num, self.flags, self.cmdID, cmdLength, numArgs]
			header_message = struct.pack(">xx2H xB 3H I 2H",*byte_array)		# Serialization per http://warpproject.org/trac/wiki/WARPLab/Reference/Architecture/WireFormat
			
			self.message = header_message + args_message
			
			#future to return when command is sent
			self.future = concurrent.futures.Future()
		
		def lastSent(self):
			"""Return how long ago this command was last sent in seconds."""
			return time.time() - self.cmdTime

		def send(self):
			"""Send the this command over the network to the node.  Increment resend counter and update time sent."""			
			self.resend_counter += 1
			self.cmdTime = time.time()

			if self.node.transport is not None:
				if self.bcast:
					self.node.transport.sendto(self.message, self.node._bcast_sock)
				else:
					self.node.transport.sendto(self.message,self.node._sock)
			else:
				self.node.logger.error("Error! Transport for %s with ID %d not initialized yet!" % (self.name, self.ID) )
					
					
	class CmdQueueManager:
		""""A class that monitors this WarpNode's command queue, resending commands if they timeout and eventually throwing an error if they fail."""

		def __init__(self, node, max_retries=10, timeout=1, check_interval=.05, max_send=1):  #the retries and timeout are a bit high, but apparently the warpnode can't keep up with python
			self.node = node
			self.max_retries = max_retries
			self.timeout = timeout
			self.check_interval = check_interval
			self.max_send = max_send  #maximum number of cmds that can be resent simultaneously.  Note that this does not limit number of commands initially sent.
		def check(self):
			"""Check the command queue for commands that have timed out and resend them or fail them."""
			delete = []	
 
			num_sent = 0 #only resend up to max_sent.
			for k in list(self.node.cmdQueue.keys()): #make copy of keys, so it doesn't change.
				cmd = self.node.cmdQueue[k]
				if cmd.finished:
					delete.append(cmd.seq_num)		#threadsafe deletion (otherwise you get RuntimeErrors)		
				elif cmd.lastSent() > self.timeout and num_sent < self.max_send:
					# if this command has been resent for more than 5 times drop it from queue and do nothing
					if cmd.resend_counter < self.max_retries:
						#print("Resending command seq %d from node %d, try: %d" % (cmd.seq_num, self.node.ID,cmd.resend_counter))  
						cmd.send()
						num_sent += 1 
						#del self.node.cmdQueue[cmd.seq_num]
					else:
						cmd.future.set_result(None)
						self.node.logger.error("Error! Node %d did not receive command seq %d, cmd %x, after %d retries! cmdQueue still has %d cmds in it." % (self.node.ID, cmd.seq_num, cmd.cmdID, cmd.resend_counter, len(self.node.cmdQueue)))
						print("Error! Node %d did not receive command seq %d after %d retries! cmdQueue still has %d cmds in it." % (self.node.ID, cmd.seq_num, cmd.resend_counter, len(self.node.cmdQueue)))
						# raise an exception when a cmd exhausts its retries
						delete.append(cmd.seq_num)
						#[print('%x' % self.node.cmdQueue[seq].cmdID) for seq in k]						
						#raise ConnectionError("Error! Node %d did not receive command seq %d, cmd %x, after %d retries! cmdQueue still has %d cmds in it." % (self.node.ID, cmd.seq_num, cmd.cmdID, cmd.resend_counter, len(self.node.cmdQueue)))
			for seq_num in delete:  #delete after so the dictionary doesn't change size during iteration
				del self.node.cmdQueue[seq_num]

		def startAutoCheck(self):
			"""Schedule the manager to check the queue at a regular interval."""
			#self.node.logger.info("AutoCheck starting...")
			self.check()
			self.node.loop.call_later(self.check_interval, self.startAutoCheck)
			#asyncio.get_event_loop().call_later(self.check_interval, self.startAutoCheck)

		def wait(self, interval=.05):
			"""Wait for the command queue to empty (meaning replies have been received for every command sent).  This is used for the barrier function."""				
			while self.node.cmdQueue:
				#if not asyncio.get_event_loop().is_running():
				#	asyncio.get_event_loop().run_until_complete(asyncio.sleep(.05))
				if not self.node.loop.is_running():  #we can probably remove this now that we are sure the event loop should *always* be running.  Previously, this was a hackish way to run the event loop temporarily to clear out the cmdQ
					self.node.loop.run_until_complete(asyncio.sleep(.05))
				else:
					#print(self.node.ID,self.node.cmdQueue)
					#print(self.node.loop.is_running())
					time.sleep(interval)
			
	def __init__(self, serial=None, ID=None, IP=None, name=None, discover_nodes=False, dont_init=False):
		"""Initialize this node with the specified ID, IP, Serial, and Name."""

		#instance variables
		self._seq_num = 0
		self._mtu =  8962 #1400	
					
		#Basic configuration from input parameters.
		self.serial = serial
		self.ID = ID		
		self.IP = IP
		self.name = name
		
		self.dummy = False
		if discover_nodes:
			self.IP = '10.224.76.100'
			self.name = 'DummyDiscoveryNode'
			self.ID = 100
			self.serial = 100
			self.dummy = True
			self.IDcnt = 0
			print('serialNumber,ID,ipAddress,name')
			
		self._base_IP = '.'.join(self.IP.split('.')[:3])+'.'
		self._bcast_addr = self._base_IP + '255'  #this only works if it is a 255.255.255 netmask #todo: make this work for all networks.
		self._bcast_sock = (self._bcast_addr,self._bcast_port)

		# create logger
		self.logger = logging.getLogger() #'warpNode.py - initializing node: %s' % str(self))
		self.logger.setLevel(logging.DEBUG)
		# create console handler and set level to debug
		if not os.path.exists('logs'):
			os.makedirs('logs')
		ch = logging.FileHandler('./logs/debuginfo.log')
		#ch = logging.StreamHandler(sys.stdout) #log to stdout
		ch.setLevel(logging.DEBUG)
		# create formatter
		formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
		# add formatter to ch
		ch.setFormatter(formatter)
		# add ch to logger
		self.logger.addHandler(ch)
		
		#TODO: reenable this (and pass config file? or perhaps just have these be optional parameters?)
		#config = configparser.ConfigParser()
		#ConfigParser().read('wl_config.ini')
		#self._host_ID = config.getint('network', 'host_ID')
		#self._host_address = config.get('network', '_host_address')
		#self._unicast_starting_port = config.getint('network', '_unicast_starting_port')
		#self._bcast_port = config.getint('network', '_bcast_port')
		#self._jumbo = config.getboolean('network', '_jumbo')
		
		#todo: Make interface an instance class? Then set this in the class?
		self.tx_gain_setting = {self.RFA : [0,0], self.RFB : [0,0], self.RFC : [0,0], self.RFD : [0,0]}  
		self.rx_gain_setting = {self.RFA : [0,0], self.RFB : [0,0], self.RFC : [0,0], self.RFD : [0,0]}
		
		IPs = self.get_local_ips()
		if len(IPs) < 1:
			raise RuntimeError('Error! No valid network interfaces (with valid IP) found!')
		if len(IPs) > 1:
			print('Warning! There is more than one network interface that matches this node''s network.  Using first one to set host ID')
		self._host_ID = int(IPs[0].split('.')[3])
		#self._host_address = '10.224.76.'+str(self._host_ID) #listen only on this interface
		self._host_address = '0.0.0.0' #listen on any interface
		
		self._sock = (self.IP,self._unicast_starting_port+self.ID)
		self._host_sock = (self._host_address,self._unicast_starting_port+self.ID)  #might as well match host port to client port.
		
		self.loop = asyncio.new_event_loop()
		self.loop.run_until_complete(self.loop.create_datagram_endpoint(lambda: self, local_addr=self._host_sock) ) #this is supposed to support , allow_broadcast=True
		self.thread_loop = threading.Thread(target=self.loop_in_thread)
		
		self.cmdQueue = {}
		self.cmdQMan = self.CmdQueueManager(self)	
		self.cmdQMan.startAutoCheck()  #the autocheck has to be scheduled before the loop starts running in a thread, apparently.
		
		self.thread_loop.start()	
		
		self.socket = self.transport.get_extra_info('socket')
		self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1) #todo: remove this, it is supposed to be supported with , allow_broadcast=True in the create_datagram_endpoint
		#self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

		#Implements "WarpNode(discover_nodes=True)" to print out a list of nodes on the network.
		if discover_nodes:
			self.sendCmd(self.calcCmd(self._GRPID_NODE,self.info), [self.serial, self.ID, self.IP2Int(self.IP,swap_endian=True)], bcast=True )
			time.sleep(1)
			self.close()
		else:
			self.ip_setup() #this shouldn't hurt even if the node isn't using dynamic config, so let's go ahead and do it here.
			if not dont_init: #temporary hack to avoid 5GHz issue (or avoid resetting PLLs)
				self.nodeCmd('initialize')			
			self.nodeCmd('info')
			self.barrier()  #make sure we have established communication before we try to find the MTU
			self._mtu = self.find_mtu()  #this is blocking, which is not good.
			self.baseband._mtu = self._mtu #override the baseband _mtu
		

	def loop_in_thread(self):
		"""A thread calls this to start the event loop to listen to datagrams on the socket."""
		asyncio.set_event_loop(self.loop)
		self.loop.run_forever()
	
	def get_local_ips(self):
		"""Return a list of IPs for the local interfaces."""
		ips = []
		for interface in nif.interfaces():
			if nif.AF_INET in nif.ifaddresses(interface):
				#print(nif.ifaddresses(interface)[nif.AF_INET])
				for i in nif.ifaddresses(interface)[nif.AF_INET]:
					ips.append(i['addr'])
		#ips = socket.gethostbyname_ex(socket.gethostname())[2]	#Lists all current IP addresses
		matching = [ip for ip in ips if self._base_IP in ip]  # Find appropriate IP address based on base IP address
		return matching #return all matching.
		
	def get_local_macip(self, IP=None):
		"""Return the (mac,ip) of this node's interface, or the (mac,ip) of a specified IP, if it is provided."""
		if IP is None:
			IPs = self.get_local_ips()
			if len(IPs) > 1:
				print('Warning! More than one network interface matches the node network.')
			IP = IPs[0]		
		
		for i in nif.interfaces():	  # Find corresponding MAC address
			addrs = nif.ifaddresses(i)
			try:
				if_mac = addrs[nif.AF_LINK][0]['addr']
				if_ip = addrs[nif.AF_INET][0]['addr']
				if str(if_ip) == IP:
					mac = str(if_mac)
					break
			except KeyError: #ignore interfaces that don't have MAC or IP
				if_mac = if_ip = None	
		if if_mac == None:
			raise Exception('Error: Can''t find current IP/MAC address')
		return (IP,mac)

		
	def nodeCmd(self, cmdID, *args, port=None, callback=None):
		"""Implements the node command -- creates a network packet with the appropriate headers and options/data and sends it using sendCmd."""
		callback = [] if callback is None else   callback if type(callback) is list else [callback]  #convert callback to a list if it isn't.
		if type(cmdID) is str:
			cmdID = getattr(self, cmdID.lower())
			
		if cmdID == self.set_cc_macip: # set_cc_macip sets which MAC/IP/Port to send IQ data to, and how much data.  takes up to 4 arguments: ip, mac, num_mobs, port.  ip and mac will automatically be determined if not provided
			if len(args) < 2:
				ip,mac = self.get_local_macip()
			else:
				ip,mac = args[0],args[1]

			#format the macip for packetization.	
			mac_split = [int(n,16) for n in ( mac.split(mac[2]) if len(mac) == 17 else list(map(''.join, zip(*[iter(mac)]*2))) )] #if the MAC has delimiters use it to split, otherwise split it in to chunks of 2, then cast each from hex to an int16.
			mac_low = mac_split[5]*2**24 + mac_split[4]*2**16 + mac_split[3]*2**8
			mac_high = mac_split[2]*2**24 + mac_split[1]*2**16 + mac_split[0]*2**8
			ip = self.IP2Int(ip)
			macip = [ip] + [mac_low] + [mac_high]
				
			if len(args) == 3:			   # If MAC and IP and all additional arguments are provided
				args = macip + args[3]
			elif len(args) <= 1:			 # MAC and IP are not provided
				args = macip + list(args)
			else:							# No additional arguments but MAC and IP are provided
				args = macip
			if port is not None and len(args) == 4:  #todo, probably just make setting num_mobs a different command
				args.append(port)  #the port has to be argument 4, so we have to set the num_mobs here
			
		if cmdID == self.info:
			callback.append(self.info_callback) # if the command is info, then add the corresponding callback to the callback list

		thiscmd = self.sendCmd(self.calcCmd(self._GRPID_NODE,cmdID), args, callback=callback)
		
		return thiscmd.future

	class CombineResultList(CallableList):
		"""A CallableList that merges the results in to a single numpy array.  Made for the read buffer operation."""
		def __init__(self, *a, **kw):
			super(WarpNode.CombineResultList, self).__init__(*a, **kw)
		def result(self):
			return np.concatenate([f.result() for f in self])
			
	def basebandCmd(self, cmdID, *args, callback=None):
		"""Implements the baseband command -- creates a network packet or packets with the appropriate headers and options/data and sends it using sendCmd."""
		callback = [] if callback is None else callback if type(callback) is list else [callback]  #convert callback to a list if it isn't.
		
		cmdarg, args_list, flag, combine_results = self.baseband.procCmd(cmdID, *args, mtu=self._mtu)	#this returns a list of args that should be sent in multiple packets	#todo pass back the args_message (and make sendCmd take args_message), as well as a callback to process putting together reads that are more than the MTU, or multiple register reads

		if args_list != [] and type(args_list[0]) == list:
			r = CallableList([self.sendCmd(self.calcCmd(self.baseband.GRPID,cmdarg), args, flags=flag, callback=callback).future for args in args_list])
			r = self.CombineResultList(r) if combine_results else r # we have to combine each of the packet results on a read buff command.
			return r[0] if len(r) == 1 else r  #hrm, not sure if I like this
		else:
			return self.sendCmd(self.calcCmd(self.baseband.GRPID,cmdarg), args_list, flags=flag, callback=callback).future   #shouldn't ever get here...

	def wsdInit(self, callback=None):
		"""This is the only WSD (WURC) command that isn't an interface command."""
		return self.sendCmd(self.calcCmd(self._GRPID_WSD,1), [], callback=callback).future  #this happens to be the same as the node 'initialize' command, but since it doesn't have to be let's explicitly used the command ID, which is 1
	
	def interfaceCmd(self, cmdID, ifaces, *args, callback=None):
		"""
		Implements the interface command -- creates a network packet with the appropriate headers and options/data and sends it using sendCmd.
		This operates on individual radio interfaces, which must be provided, either individually or in a list, or 'RFALL'.
		"""
		callback = [] if callback is None else callback if type(callback) is list else [callback]  #convert callback to a list if it isn't.
		
		#parse the targeted interfaces.  This flexibly takes things like [WarpNode.RFA,'RFB'] or WarpNode.RFA + WarpNode.RFB, then combines them all in to one bitmask.
		if type(ifaces) is str:
			ifaces = getattr(self, ifaces.upper())  #let the user pass 'rfa' as the identifier.
		if type(ifaces) is list:
			out = 0
			for i in range(len(ifaces)):
				if type(ifaces[i]) is str:
					ifaces[i] = getattr(self, ifaces[i].upper())
				out |= ifaces[i]
			ifaces = out
		
		ftrs = CallableList()
		
		#for now, assume the user knows what they are doing, and won't send commands for maxim to wurc and vice versa; just send the command as described.
		if ifaces & self.RFALL_M2829:
			cmdarg, sendargs = InterfaceM2829.procCmd(cmdID, *args) #[i]+args[1:])
			sendargs = [ifaces & self.RFALL_M2829] + sendargs
			ftrs.append(self.sendCmd(self.calcCmd(self._GRPID_IFC,cmdarg), sendargs, callback=callback).future)
		if ifaces & self.RFALL_LMS6:
			cmdarg, sendargs = InterfaceLMS6.procCmd(cmdID, *args) #[i]+args[1:])
			sendargs = [ifaces & self.RFALL_LMS6] + sendargs
			ftrs.append(self.sendCmd(self.calcCmd(self._GRPID_WSD,cmdarg), sendargs, callback=callback).future)
			#self.barrier() #apparently the wsd commands are very slow, so let's wait for them to complete before we continue...
			
		#temporary hack to save gains for uplink AGC.  TODO: implement an increase gain command on the mobile
		#update gains.
		
		if cmdID == 'tx_gains': #cmdID == InterfaceM2829.tx_gains or 
			for r in self.RFS:
				if ifaces & r:
					self.tx_gain_setting[r][:] = args #[:] so that we don't create a new list
		elif cmdID == 'rx_gains': #cmdID == InterfaceM2829.rx_gains or 
			for r in self.RFS:
				if ifaces & r:
					self.rx_gain_setting[r][:] = args #[args[0],args[1]]
		
			
		return ftrs
		
		
	def sdCmd(self, cmdID, *args, callback=None):
		"""
		Implements the sdCmd command for writing SD cards and rebooting nodes.
		In WARPLab this is implemented in the user_ext class.
		Creates a network packet or packets with the appropriate headers and options/data and sends it using sendCmd.
		"""
		callback = [] if callback is None else   callback if type(callback) is list else [callback]  #convert callback to a list if it isn't.
		
		cmdarg, args = sd_cfg.procCmd(self, cmdID, *args, mtu=self._mtu)		
		args = [int(n) if isinstance(n,float) else n for n in args]  # Converts arguments to int if need be		
		
		if args != [] and type(args[0]) == list:
			num_pkts = len(args)
			p = 0
			if cmdID == 'sd_write_bitstream':  
				print('Writing SD Image: 00%%')
				r = CallableList()
				for arg in args:
					r.append(self.sendCmd(self.calcCmd(self._GRPID_USER,cmdarg), arg, flags=1, callback=callback).future)
					print('\b\b\b%02d%%' %np.floor(100*p/num_pkts))
					ret = r.result()
					#print (ret)
					#if(ret != 0):
						#print('SD write failed on pkt %d!\n' %p) 
						#return
					p +=1
			print('Done\n\n')
		else:
			return self.sendCmd(self.calcCmd(self._GRPID_USER,cmdarg), args, flags=1, callback=callback).future
		
		
	def __del__(self):
		self.transport.close()
		del self
		
	def __str__(self):		
		typestr = "NA" if self.nodetype is None else self._TYPE_NAMES[self.nodetype]
		serial = "0" if self.serial is None else self.serial
		return " %3s | W3-a-%05d | %2d | %16s | %16s |\n" % (typestr, serial, self.ID, self.IP, self.name)

	def __repr__(self):
		return self.__str__()	

	@staticmethod
	def table_header(cls):
		return(' Type  |   Serial   | ID |        IP        |       Name       | ')
	
	@staticmethod
	def IP2Int(ip, swap_endian=False):
		"""Convert an IP string in to a 4 byte integer."""
		#TODO: make ip_setup and set_cc_macip use the same endianness...
		o = list(map(int, ip.split('.')))		
		return (o[0]<<24) + (o[1]<<16) + (o[2]<<8) + o[3] if swap_endian else (o[3]<<24) + (o[2]<<16) + (o[1]<<8) + o[0]
	
	@staticmethod
	def eth2str(eth):
		return ('%x-%x-%x-%x-%x-%x' % ( (eth >> 40 ) & 0xFF, (eth >> 32 ) & 0xFF, (eth >> 24 ) & 0xFF, (eth >> 16 ) & 0xFF, (eth >> 8 ) & 0xFF, (eth >> 0 ) & 0xFF ) )
		
	#let's keep this as a separate command since it uses all internal values, as well as requires broadcast
	def ip_setup(self):
		"""
		Set a node's IP using the WARP protocol (broadcast a packet with a serial # and ID)

		The node must be configured to use a dynamic IP, all dipswitches up. 
		(Or this must match it's dipswitch config.)
		"""
		ipCmd = self.sendCmd(self.calcCmd(self._GRPID_NODE,self.setup), [self.serial, self.ID, self.IP2Int(self.IP,swap_endian=True)], bcast=True )
		#del self.cmdQueue[ipCmd.seq_num]  #this doesn't work always! (command is dequeued before this is called) instead we just don't ever queue broadcast commands.  #dequeue this from the cmdQ since Rice Wireless blocks broadcast packets.  This could be just mobiles... but no reason we have to have it for other nodes.

	
	def connection_made(self, transport):
		self.logger.info('Server Started for %s with ID %d' % (self.name, self.ID) )
		self.transport = transport

	def datagram_received(self,data,addr):
		"""
		Function that handles all packets for this node.
		First it parses and unpacks all header info and data.
		Then it sets the result of the future for the command and removes it from the command queue.
		Finally, it runs any callbacks that are set.
		This also handles auto-discovery of nodes, printing them out as	broadcast replies are received.
		"""
		#print("Data received from %s for node %d" % (addr[0], self.ID) )
		#self.logger.info("Data received from %s for node %d" % (addr[0], self.ID) )
		header = struct.unpack(">xx2H xB 3H I 2H", data[0:22])  #> means big endian.  H is 2 bytes, B is 1, I is 3, x is disregarded single pad byte.  capital is unsigned
		ID = header[0]
		nodeID = header[1]
		pkt_type = header[2]
		length = header[3]
		seq_num = header[4]
		flags = header[5] 
		cmd = header[6]
		cmdGrp = cmd>>24
		cmdID = cmd & 0xFFFFFF
		cmdLength = header[7]
		numArgs = header[8]

		#cmdArgs = struct.unpack(">%dI" % ( (len(data)-22)//4 ), data[22:])
		if self.baseband is not None and cmdGrp == self.baseband.GRPID: #(cmdGrp == _GRPID_BB or cmdGrp == _GRPID_ARGOSBS or cmdGrp == _GRPID_ARGOSMOBILE) and self.baseband is not None:
			cmdArgs = self.baseband.procReply(cmdID,data)
		else:
			cmdArgs = struct.unpack(">%dI" % ( (len(data)-22)//4 ), data[22:])
		
		#print ("Received from Node %d, IP %s, seqnum %d, cmd %x (cmdID %x, cmdGrp %x), this node is %d, cmdLength %d, numArgs %d, CmdQueueManager %s" % (nodeID, addr, seq_num, cmd, cmdID, cmdGrp, self.ID, cmdLength, numArgs, self.cmdQMan))
		#for cmds in self.cmdQueue:
			#print ("\tCommands in Command Queue", self.cmdQueue[cmds].seq_num, self.cmdQueue[cmds].cmdID)
	
		#self.logger.info("Packet from node %x with cmd %x (cmdID %d, cmdGrp %d) and %d cmdArgs received!" % (nodeID, cmd, cmdID, cmdGrp, numArgs) )
		
		# set future's result, process callbacks, and delete the completed command from the command queue
		if seq_num in self.cmdQueue:
			self.cmdQueue[seq_num].future.set_result(cmdArgs)
			# execute the callback function for this command
			
			for cb in self.cmdQueue[seq_num].callback:
				cb(cmdArgs)  #maybe this should just be cb(data) or cb(data[22:])?  Or, we could use the procReply construct as above to parse out the cmdArgs, which is probably better.

			#set this command as finished
			self.cmdQueue[seq_num].result = list(cmdArgs)
			self.cmdQueue[seq_num].finished = True  # remove this command from the command Queue #we have to let the CmdQueueManager remove it in order to be threadsafe (you can get random RuntimeError: dictionary changed size during iteration errors otherwise)
			
			#self.logger.info("Packet from node %x with cmd %d (cmdID %d, cmdGrp %d) and %d cmdArgs received!" % (nodeID, cmd, cmdID, cmdGrp, numArgs) )
		
		if cmdGrp == self._GRPID_NODE and cmdID == self.info and len(cmdArgs) >= 10 and self.dummy:
			"""For auto-discovery of nodes."""
			self.IDcnt += 1
			nodetype = cmdArgs[10] if len(cmdArgs) >= 11 else 0
			if nodeID > 15 and nodeID != 0xFFFF:
				print('Warning! Detected nodeID greater than 15, which means it was probably already configured dynamically.')

			ips = addr[0].split('.')
			ip = int(ips[3])
			if nodeID == 0xFFFF: #ip == 100:
				nodeID = self.IDcnt+20
				if nodetype & self.ARGOSBS_TYPE:		
					ip = nodeID+101
				elif nodetype & self.ARGOSMOBILE_TYPE:
					ip = nodeID+201
				else:
					ip = nodeID+1 #WARPLab of some sort, probably not a valid IP.

			print("W3-a-%05d,%d,%d.%d.%d.%d,%s_%d" % (cmdArgs[0], nodeID, int(ips[0]), int(ips[1]), int(ips[2]), ip, self._TYPE_NAMES_LONG[nodetype], self.IDcnt))
			
		elif nodeID != self.ID:
			if cmdID != self.setup: #all nodes respond to a setup command
				self.logger.warn("Warning! Received packet from another node (which shouldn't happen in current architecture)! %d ")  #apparently there are computers on rice research that respond to broadcast packets?
			return
			
			
	def info_callback(self, cmdArgs):
		"""
		This is the callback that processes the reply from the info command.  
		It sets all of the configuration (v3, software version, baseband, interface, etc.).
		"""
		self.logger.info("Received hardware info back for node: %s!" % str(self) )
		numArgs = len(cmdArgs)
		if len(cmdArgs) < 10 or numArgs < 10:
			self.logger.warn("Warning!  Bad info returned from node!")
			return
		if self.serial is not None and self.serial != cmdArgs[0]:
			self.logger.warn("Warning! Serial returned from info cmd does not match serial!")
		self.serial = cmdArgs[0]
		#print("Received info call back for node %d packet identifies node serial %s." % (self.ID, cmdArgs[0]))
	
		self.fpgaDNA = cmdArgs[1]<<32 + cmdArgs[2]			
		self.eth_MAC_addr = (( cmdArgs[3] & 0xFFFF ) << 32 ) + cmdArgs[4]
	
		self.hwVer = (cmdArgs[5] >> 24 ) & 0xFF
		self.wlVer_major = (cmdArgs[5] >> 16 ) & 0xFF
		self.wlVer_minor = (cmdArgs[5] >> 8 ) & 0xFF
		self.wlVer_revision = cmdArgs[5] & 0xFF
	
		self.interfaceGroups = cmdArgs[8]
		self.num_m2829 = cmdArgs[9] & 0xFF
		self.num_lms6 = (cmdArgs[9] & 0xFF00) >> 8
		self.num_interfaces = self.num_m2829 + self.num_lms6
		self.RFALL = sum(self.RFS[0:self.num_interfaces])
		self.RFALL_LMS6 = sum(self.RFS[0:self.num_lms6])  #we assume the WSD interfaces are RFA, RFB, RFC, etc., and the m2829 interfaces come after the wsd interfaces
		self.RFALL_M2829 = self.RFALL - self.RFALL_LMS6
		if self.RFALL_LMS6 > 0:
			self.wsdInit()  #hmm, I'm not sure how much I like one command calling another, but this should be fine.	
		self.IFCS = {}
		for rf in self.RFS:
			self.IFCS[rf] = InterfaceM2829 if rf & self.RFALL_M2829 else InterfaceLMS6

		if len(cmdArgs) == 10:
			self.nodetype = self.WARPLAB_TYPE
			self.baseband = wl_baseband()   
			
		elif len(cmdArgs) >= 11:
			self.nodetype = cmdArgs[10]
			if self.nodetype & self.ARGOSBS_TYPE: #mask since WURC nodes report 5 (bs) and 6(mobile)
				if self.nodetype & self.WURC_TYPE: #separate WURC baseband since it is substantially different now (no AGC, no RSSI, and has multirate)
					self.baseband = ArgosBS_WURC()
				else:
					self.baseband = ArgosBS()
				self.cmdQMan.timeout = 1 #should be .1, this is a hack for the wurc init which takes a while.  TODO: set timeout just for WURC.  #by default we have a very large and long timeout to accomodate the mobiles, but we don't want to wait 5s for an error always.
			elif self.nodetype & self.ARGOSMOBILE_TYPE:
				self.baseband = ArgosMobile()
				self.cmdQMan.max_retries = 25  #when we're using wireless bridges we want it to take a bit longer to time out
				self.cmdQMan.timeout = 1
				self.cmdQMan.max_send = 1
				#self._mtu = 1400  #only do this if not set automatically
				#print('Warning! Mobile node mtu still set to %d!' % self._mtu)
			elif self.nodetype == self.WARPLAB_TYPE:  #should never get here, by default WL doesn't report a type.
				self.baseband = wl_baseband()
				
		if len(cmdArgs) >= 12:
			self.version_hash = cmdArgs[11]
			
		if len(cmdArgs) > 12:
			self.logger.warn('Warning! Info cmd returned more than 12 arguments for node %d!' % self.ID)
	
		self.baseband._mtu = self._mtu #override the baseband _mtu
		typestr = "NA" if self.nodetype is None else self._TYPE_NAMES_LONG[self.nodetype]
		self.description = 'Initialized %s node with ID %d named %s.' % (typestr, self.ID, self.name) 
		self.inited = True
		self.logger.info('Received info for node %d, %s.  It has serial number W3-a-%05d, MAC address %s, is running %s %d.%d.%d, and has %d interfaces.' % (self.ID, self.name, self.serial, self.eth2str(self.eth_MAC_addr), typestr, self.wlVer_major, self.wlVer_minor, self.wlVer_revision, self.num_interfaces) )


	def connection_refused(self, exc):
		self.logger.info('Connection refused: %s' % exc)

	def connection_lost(self, exc):
		self.logger.info('Connection lost %s' % exc)	
	
	def sendCmd(self, cmd, cmdArgs=[], bcast=False, flags=1, callback=None):
		"""Create a new command object with the next seq_num, put it in the command queue, and send it."""
		if callback is None:
			callback = []
		self._seq_num = (self._seq_num + 1) & 0xFFFF #have to mask as uint16, otherwise it overflows in the struct.pack
		new_cmd = self.WarpCommand(self, self._seq_num, cmd, args=cmdArgs, flags=flags, bcast=bcast, callback=callback)
		if not bcast: #don't require a reply from broadcasts.  If they didn't setup their IP, they will error on the next command, and currently the rice network breaks most broadcast.
			self.cmdQueue[self._seq_num] = new_cmd
		new_cmd.send()
		return new_cmd
	
	@staticmethod
	def calcCmd(group, cmd):
		return (group<<24)+cmd #really we should check if they are valid...
	

	def barrier(self):
		"""Wait to receive a reply (ack) for all pending commands by waiting for the command queue to empty.  If called with WarpNode.barrier(nodes), then barrier on all of the nodes."""
		[n.cmdQMan.wait() for n in (self if isinstance(self, list) else [self])]  #this allows it to be called via WarpNode.barrier(listofnodes) or node.barrier()

			
	def error_received(self, arg):
		self.logger.info('Error received %s!' % str(arg))
		return 

	#The stackoverflow solution:
	#http://stackoverflow.com/questions/14010635/how-to-find-mtu-value-of-network-through-codein-python
	#It's also outdated, so we change a few fields so that everything should work.
	#Choice of port should be unimportant -- as long as the packet isn't dropped by a firewall,
	#we should have luck.
	#this appears to be linux only...
	'''
	def find_mtu(self, host=None, port=7777):				
		host = host if host is not None else self.IP
		sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		sock.connect((host, port))
		#Make a garbage buffer 60,000 bytes in size. We should not be able to send thisin one go.
		sock_buf = bytearray(b'x' * 60000) 
		sock.setsockopt(socket.IPPROTO_IP, IN.IP_MTU_DISCOVER, IN.IP_PMTUDISC_DO)
		try:
			#This should fail because the MTU is exceeded
			sock.send(sock_buf)
		except socket.error:
			option = getattr(IN, 'IP_MTU', 14)
			mtu = sock.getsockopt(socket.IPPROTO_IP, option)
		else:
			mtu = 60000
			raise Warning("MTU exceeds 60,000. This is very unlikely.")
		return mtu
	'''

	
	def find_mtu(self, callback=None):
		"""
		Send packets of increasing size until they fail to get the empirical MTU.
		Note that while a single large packet will suffice on most networks 
		(since the node will still return how many bytes it received),
		some networks *discard* large packets, hence we need to loop.
		
		Note that we also only try once, to avoid long delays for finding MTUs.
		"""
		callback = [] if callback is None else callback if type(callback) is list else [callback]  #convert callback to a list if it isn't.

		payloadTestSizes = (np.array(([1000, 1470, 5000, 8966])) - (12 + 8) - 4)//4  #header takes up space
		cmd = []
		for index in range(len(payloadTestSizes)):
			cmd.append(self.sendCmd(self.calcCmd(self._GRPID_TRANS,4), [x for x in range(1, payloadTestSizes[index]+1)], 0, callback=callback)) #.future.result()
			cmd[index].resend_counter = self.cmdQMan.max_retries - 1
		for index in range(len(payloadTestSizes)):
			try:
				resp = cmd[index].future.result()
			except ConnectionError:
				print ("Failed to get response for node %d packet of size: %d" % (self.ID, str(payloadTestSizes[index]*4)) )
				break
				
			if type(resp) != None:
				try:
					mtu = resp[0]
					if resp < payloadTestSizes[index]*4:
						break
				except TypeError:
					if payloadTestSizes[index]*4 < 2000:
						print ("Failed to get response for node %d packet of size: %d" % (self.ID, str(payloadTestSizes[index]*4)) )
						print(self)
					break

		return mtu
			
	
	
	def close(self):
		"""Stop the transport, stop the loop, and close the loop.  Take care of any cleanup."""
		self.transport.close()
		self.loop.call_soon_threadsafe(self.loop.stop) #self.loop.stop() is not threadsafe.
		while self.loop.is_running():
			time.sleep(.05)
		self.loop.close()	
		self.thread_loop.join() #probably not necessary, but good form.
		
	def reboot(self, slot=None):
		"""Reboot node to specified slot.  If no slot is specified, it will guess based on the current node type."""
		self.sdCmd('sd_init')
		if slot is None:
			slot = 0 if self.nodetype == self.WARPLAB_TYPE else 1				
		self.sdCmd('sd_reconfig', slot)

#Example code for testing and verification -- everything below here is not part of warpNode.py		
import sys
import code

if __name__ == '__main__':
	nodes = initNodes(nodesFile='./argos.ips')
	print(nodes)
	#[n.nodeCmd('identify') for n in nodes] # or nodes.nodeCmd('identify')
	code.interact(local=locals()) 
